/***************************************************************************
 *                                                                         *
 *                  (begin: Feb 20 2003)                                   *
 *                                                                         *
 *   Parallel IQPNNI - Important Quartet Puzzle with NNI                   *
 *                                                                         *
 *   Copyright (C) 2005 by Le Sy Vinh, Bui Quang Minh, Arndt von Haeseler  *
 *   Copyright (C) 2003-2004 by Le Sy Vinh, Arndt von Haeseler             *
 *   {vinh,minh}@cs.uni-duesseldorf.de                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#ifndef IQP_H
#define IQP_H

#include <ctime>
#include "constant.h"
#include "vec.h"
#include "urtree.h"
#include "treels.h"
#include "clusterarr.h"

/**
	Important Quartet Puzzling Class
*/
class IQP {
public :
	int nSeq_;
	int nRep_;
	double probDel_;
	ClusterArr bestClusterArr_;


	/**
	we will init all things here before starting testing this method with data
	*/
	void init ();

	/**
		optimize parameters at final step
	*/
	double tuneFineParameter (int nIter, UrTree &bestTree);

	void optPam ();

	double escape (Vec<int> &orderNdNoLs, UrTree &tree, UrTree &bestTree, ClusterArr &bestClusterArr);

	double escapeLocalOpt (int &nIt, int max_nIter, double proSel, int nRep, STOPPING_TYPE stoppingRule, 
		UrTree &bestTree, const char *userTreeFile, int outGrpSeqNo, Vec<char> &conTree,  
		time_t &beginTime_, double &progTime);

	void        loadTreeLs (const char *treeLsFileName, int &nIter, ClusterArr &bestClusterArr, ClusterArr &allCluArr,
	                        UrTree &bestTree, double &bestLogLi,
	                        int &nImpTree, const char *predictionFileName, double &progTime);

	void writePamCheckPoint ( char  *checkPointFileName,
	                          int        nIter,
	                          int        max_nIter,
	                          double     proSel,
	                          int        nRep,
	                          STOPPING_TYPE     stoppingRule,
	                          int         outGrpSeqNo,
	                          int         isFinished);
}; //end of class;

#endif

