/***************************************************************************
 *                                                                         *
 *                  (begin: Feb 20 2003)                                   *
 *                                                                         *
 *   Parallel IQPNNI - Important Quartet Puzzle with NNI                   *
 *                                                                         *
 *   Copyright (C) 2005 by Le Sy Vinh, Bui Quang Minh, Arndt von Haeseler  *
 *   Copyright (C) 2003-2004 by Le Sy Vinh, Arndt von Haeseler             *
 *   {vinh,minh}@cs.uni-duesseldorf.de                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

/***************************************************************************
                          libr.cpp  -  description
                             -------------------
    begin                : Thu Apr 24 2003
    copyright            : (C) 2003 by 
    email                : vinh@cs.uni-duesseldorf.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <math.h>


#include <time.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "constant.h"
#include "libr.h"
#include "ptnls.h"
#include "rate.h"
#include "model.h"
#include "opturtree.h"
#include "dmat20.h"
#include "ptnratecube.h"

#ifdef PARALLEL
#include <mpi.h>
#endif
//extern int nto[NUM_CHAR];

#ifdef WIN32
#include <float.h>
#define isnan(x) _isnan(x)
#endif

const double VERY_SMALL_POSITIVE = 1e-200;

//#define PRINT_NEGATIVE_INFO

	/**
		the constructor
	*/
LiBr::LiBr () {
	smaLiNd_ = new LiNd;
	greLiNd_ = new LiNd;
}

/*
#if !defined(_isnanl128)
int _isnanl128(LDOUBLE x) {
	LDOUBLE y = x;
	return (x != y);
}
#endif
*/

	/**
		copy the base br into this likelihood branch
	*/
void LiBr::copyBase (Br<double> &br) {
	this->id_ = br.getId ();
	this->smaNdNo_ = br.getSmaNd ();
	this->greNdNo_ = br.getGreNd ();
	this->isEx_ = br.isEx ();
	this->len_ = br.getLen ();
	this->isStG_ = br.isStG ();

	greLiNd_->set (greNdNo_, IN, smaNdNo_, id_);
	if (isEx_ == 1)
		smaLiNd_->set (smaNdNo_, EX, greNdNo_, id_);
	else
		smaLiNd_->set (smaNdNo_, IN, greNdNo_, id_);
}


	/**
		@return 1 if this is an outgroup node
	*/
int LiBr::isOutGrp () {
	if (id_ == opt_urtree.outGrpBrNo_ )
		return 1;
	else
		return 0;
}

	/**
		@param existNdNo the node ID at the opposite
		@return the remaining liNd of this branch 
		opposite to existNdNo
	*/
LiNd &LiBr::getRemLiNd (int existNdNo) {
	if (existNdNo != smaNdNo_)
		return *smaLiNd_;
	else
		return *greLiNd_;
}

	/**
		@return the small liNd of this branch
	*/
LiNd &LiBr::getSmaLiNd () {
	return *smaLiNd_;
}

	/**
		@return the great liNd of this branch
	*/
LiNd &LiBr::getGreLiNd () {
	return *greLiNd_;
}

	/**
		print information 
	*/

void LiBr::writeInf () {
	std::cout << id_ << " / " << smaNdNo_ << " / " << greNdNo_ << " / " << isStG_ << endl;
	getHeadLiNd ().writeInf ();
	getTailLiNd ().writeInf ();
}
	/**
		set the new tail LiNd
		@param newTailLiNd new tail LiNd
	*/

void LiBr::setTailLiNd (LiNd *newTailLiNd) {
	if (isStG_ == 1)
		greLiNd_ = newTailLiNd;
	else
		smaLiNd_ = newTailLiNd;
	int tailNdNo_ = newTailLiNd->getId ();

	isEx_ = isExNd (tailNdNo_);

	if (smaLiNd_->getId () > greLiNd_->getId () )
		Utl::swap (smaLiNd_, greLiNd_);
	smaNdNo_ = smaLiNd_->getId ();
	greNdNo_ = greLiNd_->getId ();
	if (greNdNo_ == tailNdNo_)
		isStG_ = 1;
	else
		isStG_ = 0;
	smaLiNd_->setParNdBr (greNdNo_, id_);
	greLiNd_->setParNdBr (smaNdNo_, id_);
}

	/**
		@return the tail liNd of this branch
	*/
LiNd &LiBr::getTailLiNd () {

	if (isStG_ == 1)
		return *greLiNd_;
	else
		return *smaLiNd_;
}

	/**
		@return the head liNd of this branch
	*/
LiNd &LiBr::getHeadLiNd () {
	if (isStG_ == 1)
		return *smaLiNd_;
	else
		return *greLiNd_;
}

	/**
		@param ndNo node ID
		@return the LiNd corresponding to ndNo
	*/
LiNd &LiBr::getLiNd (int ndNo) {
	if (ndNo == smaNdNo_)
		return *smaLiNd_;
	else
		return *greLiNd_;
}

	/**
		compute the likelihood for liNd of this branch
		@param ndNo node ID
	*/
void LiBr::cmpLiNd (int ndNo) {

	if (ndNo == greNdNo_)
		greLiNd_->cmpLi ();

	if (ndNo == smaNdNo_)
		smaLiNd_->cmpLi ();
}

	/**
		compute the likelihood for tail liNd of this branch
	*/
void LiBr::cmpTailLiNd () {
	if (getTailNd () == greNdNo_) //the case, the tail node is greNdNo_
		greLiNd_->cmpLi ();
	else // the case the tail node is small node
		smaLiNd_->cmpLi ();
}

	/**
		compute the likelihood for head liNd of this branch
	*/
void LiBr::cmpHeadLiNd () {
	if (getHeadNd () == greNdNo_) //the case, the head node is greNdNo_
		greLiNd_->cmpLi ();
	else // the case the head node is small node
		smaLiNd_->cmpLi ();

}

	/**
		reCompute the likelihood for tail liNd of this branch
	*/

void LiBr::reCmpTailLiNd () {
	if (getTailNd () == greNdNo_) //the case, the tail node is greNdNo_
		greLiNd_->reCmpLi ();
	else
		smaLiNd_->reCmpLi ();
}

	/**
		reCompute the likelihood for head liNd of this branch
	*/
void LiBr::reCmpHeadLiNd () {
	if (getHeadNd () == greNdNo_) //the case, the head node is greNdNo_
		greLiNd_->reCmpLi ();
	else
		smaLiNd_->reCmpLi ();
}

	/**
		create the external descendant for all liNds of this branch
	*/
void LiBr::createExDesLiNd () {
	greLiNd_->createExDesNd ();
	smaLiNd_->createExDesNd ();
}


extern "C"
inline double cmpLogLi (LDOUBLE li) {
#ifdef USE_LONG_DOUBLE
	return logl(li);
#else
	return log(li);
#endif
}

/*	
#ifdef USE_LONG_DOUBLE
	li = logl(li);
#else
	li = log(li);
#endif
	if (isnan(li)) {
		Utl::announceError("NaN in logl");
	}
	return (double) li;
//#endif
}
*/


/***********************************************************************
* compute the probalibity change matrix times base frequency
***********************************************************************/


inline void LiBr::cmpProbChangeFrq (const double brLen, DMat20 &probMat) {
	mymodel.cmpProbChange(brLen, probMat);
	int nState_ = mymodel.getNState ();
	for (int stateNo1 = 0; stateNo1 < nState_; stateNo1 ++)
		for (int stateNo2 = 0; stateNo2 < nState_; stateNo2 ++)
			probMat[stateNo1][stateNo2] *= mymodel.stateFrqArr_[stateNo1];
}

inline void LiBr::cmpProbChangeFrq (const double brLen, const int classNo, DMat20 &probMat) {
	mymodel.cmpProbChange(brLen, classNo, probMat);
	int nState_ = mymodel.getNState ();
	for (int stateNo1 = 0; stateNo1 < nState_; stateNo1 ++)
		for (int stateNo2 = 0; stateNo2 < nState_; stateNo2 ++) {
			if (isnan(probMat[stateNo1][stateNo2]) /* != probMat[stateNo1][stateNo2]*/ ) {
				cout << "NaN" << endl;
			}
			probMat[stateNo1][stateNo2] *= mymodel.stateFrqArr_[stateNo1];
		}
}


inline void LiBr::cmpProbChangeDerivativesFrq (const double brLen, DMat20 &probMat,
        DMat20 &derv1, DMat20 &derv2) {
	mymodel.cmpProbChangeDerivatives(brLen, probMat, derv1, derv2);
	int nState_ = mymodel.getNState ();
	for (int stateNo1 = 0; stateNo1 < nState_; stateNo1 ++)
		for (int stateNo2 = 0; stateNo2 < nState_; stateNo2 ++) {
			probMat[stateNo1][stateNo2] *= mymodel.stateFrqArr_[stateNo1];
			derv1[stateNo1][stateNo2] *= mymodel.stateFrqArr_[stateNo1];
			derv2[stateNo1][stateNo2] *= mymodel.stateFrqArr_[stateNo1];
		}
}
	/**
		compute the array of probability of changing matrix 
		and the 1st and 2nd derivatives
		times frequencies between all pairs of states
		after a period of brLen / subRate_
		@param brLen branch length (or evolutional time)
		@param probMat (OUT) the prob. change matrix
		@param derv1 (OUT) 1st of P(t)
		@param derv2 (OUT) 2nd of P(t)
	*/
void LiBr::cmpProbChangeDerivativesFrq (const double brLen, int classNo, DMat20 &probMat,
                                        DMat20 &derv1, DMat20 &derv2) {
	mymodel.cmpProbChangeDerivatives(brLen, classNo, probMat, derv1, derv2);
	int nState_ = mymodel.getNState ();
	for (int stateNo1 = 0; stateNo1 < nState_; stateNo1 ++)
		for (int stateNo2 = 0; stateNo2 < nState_; stateNo2 ++) {
			probMat[stateNo1][stateNo2] *= mymodel.stateFrqArr_[stateNo1];
			derv1[stateNo1][stateNo2] *= mymodel.stateFrqArr_[stateNo1];
			derv2[stateNo1][stateNo2] *= mymodel.stateFrqArr_[stateNo1];
		}
}


/**********************************************************
***********************************************************
*    start codes to compute log likelihood and derivatives
***********************************************************
**********************************************************/

DMat20 derv1Mat, derv2Mat, proChangeMat_;

/**********************************************************
*
*    UNIFORM + INTERNAL branch
*
**********************************************************/

//--------------------------------------------------------------------
	/**
		compute the log likelihood of this tree 
		using uniform rate
		via this internal branch
		@return tree log-likelihood
	*/
double LiBr::cmpInLogLiUniformRate () {
	opt_urtree.nCmpInLogLi_ ++;

	LiNd *headLi_;
	LiNd *tailLi_;

	headLi_ = &getHeadLiNd ();
	tailLi_ = &getTailLiNd ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	//DMat20 proChangeMat_;
	cmpProbChangeFrq (len_, proChangeMat_);
	int nState_ = mymodel.getNState ();

	double inLogLi_ = 0.0;
	int nPtn_ = ptnlist.getNPtn ();
	
	#pragma omp parallel for  reduction(+: inLogLi_)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
		LDOUBLE ptnLi_ = 0.0;
		int start_addr = ptnNo_ * nState_;
		for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) 
		if (headLi_->liNdCube_[start_addr + headStateNo_] != 0.0)
		{

			LDOUBLE stateLi_ = 0.0;
			for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++)
				stateLi_ += //headLi_->liNdCube_[ptnNo_][0][headStateNo_] *
				    tailLi_->liNdCube_[start_addr + tailStateNo_] *
				    proChangeMat_[headStateNo_][tailStateNo_];
			ptnLi_ += stateLi_ * //mymodel.stateFrqArr_[headStateNo_] *
			          headLi_->liNdCube_[start_addr + headStateNo_];
		}

		//cout << "inlen=" << len_ << " li= " << ptnLi_ << endl;

		if (ptnLi_ <= 0.0) {
			ptnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double logPtnLi_ = cmpLogLi (ptnLi_);
		//logPtnLi_ += headLi_->liscale[0][ptnNo_] + tailLi_->liscale[0][ptnNo_];

		inLogLi_ += logPtnLi_ * ptnlist.weightArr_[ptnNo_];
	}

	inLogLi_ += headLi_->liscale + tailLi_->liscale;

	return inLogLi_;
}


	/**
		compute the log likelihood derivative 
		using uniform rate
		via internal branch
		@param calc_logli true if you want to compute log-likelihood
		@param logli_derv1 (OUT) 1st derivative of log-likelihood
		@param logli_derv2 (OUT) 2nd derivative of log-likelihood
		@return log-likelihood if calc_logli==true
	*/
double LiBr::cmpInLogLiDerivativeUniformRate (double &logli_derv1_ret, double
&logli_derv2_ret, bool calc_logli) {
	//opt_urtree.nCmpInLogLi_ ++;

	LiNd *headLi_;
	LiNd *tailLi_;

	headLi_ = &getHeadLiNd ();
	tailLi_ = &getTailLiNd ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	//mymodel.cmpProbChange (len_, proChangeMat_);

	cmpProbChangeDerivativesFrq (len_, proChangeMat_, derv1Mat, derv2Mat);

	int nState_ = mymodel.getNState ();
	double logli = 0.0;
	double logli_derv1 = 0.0;
	double logli_derv2 = 0.0;
	int nPtn_ = ptnlist.getNPtn ();
	#pragma omp parallel for  \
		reduction(+: logli, logli_derv1, logli_derv2)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
		LDOUBLE ptnLi_ = 0.0;
		LDOUBLE ptnLi_1 = 0.0;
		LDOUBLE ptnLi_2 = 0.0;
		int start_addr = ptnNo_ * nState_;

		for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) 
		if (headLi_->liNdCube_[start_addr + headStateNo_] != 0.0)
		{

			LDOUBLE stateLi_ = 0.0;
			LDOUBLE stateLi_1 = 0.0;
			LDOUBLE stateLi_2 = 0.0;
			for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++) {
				stateLi_ += //headLi_->liNdCube_[ptnNo_][0][headStateNo_] *
				    proChangeMat_[headStateNo_][tailStateNo_] *
				    tailLi_->liNdCube_[start_addr + tailStateNo_];
				stateLi_1 += //headLi_->liNdCube_[ptnNo_][0][headStateNo_] *
				    derv1Mat[headStateNo_][tailStateNo_] *
				    tailLi_->liNdCube_[start_addr + tailStateNo_];
				stateLi_2 += //headLi_->liNdCube_[ptnNo_][0][headStateNo_] *
				    derv2Mat[headStateNo_][tailStateNo_] *
				    tailLi_->liNdCube_[start_addr + tailStateNo_];
			}

			LDOUBLE temp = //mymodel.stateFrqArr_[headStateNo_] *
			    headLi_->liNdCube_[start_addr + headStateNo_];
			ptnLi_ += stateLi_ * temp;
			ptnLi_1 += stateLi_1 * temp;
			ptnLi_2 += stateLi_2 * temp;
		}

		if (ptnLi_ <= 0.0) {
			ptnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}
			
		double temp1 = ptnLi_1 / ptnLi_;
		double temp2 = ptnLi_2 / ptnLi_;
		if (isnan(temp1) || isnan(temp2)) {
			Utl::announceError("NaN in cmpInLogLiDerivativeUniformRate");
		}
		logli_derv1 += temp1 * ptnlist.weightArr_[ptnNo_];
		logli_derv2 += (temp2 - temp1 * temp1) * ptnlist.weightArr_[ptnNo_];
		if (calc_logli) {
			double logPtnLi_ = cmpLogLi (ptnLi_);
			logli += logPtnLi_ * ptnlist.weightArr_[ptnNo_];
		}
	}
	logli_derv1_ret = logli_derv1;
	logli_derv2_ret = logli_derv2;
	logli += headLi_->liscale + tailLi_->liscale;
	return logli;
}

/*
#ifdef PARALLEL
//--------------------------------------------------------------------
// parallel compute the likelihood of the tree, the case this branch is an internal branch
double LiBr::par_cmpInLogLiUniformRate () {
	opt_urtree.nCmpInLogLi_ ++;

	LiNd *headLi_;
	LiNd *tailLi_;

	headLi_ = &getHeadLiNd ();
	tailLi_ = &getTailLiNd ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	cmpProbChangeFrq (len_, proChangeMat_);
	int nState_ = mymodel.getNState ();

	double inLogLi_ = 0.0;
	double global_inLogLi;
	int nPtn_ = ptnlist.getNPtn ();

	int work_load = (int) ceil((double)nPtn_ / mpi_size);
	int work_start = work_load * mpi_myrank;
	int work_end = work_start + work_load;
	if (work_end > nPtn_) work_end = nPtn_;

	for (int ptnNo_ = work_start; ptnNo_ < work_end ; ptnNo_ ++) {
		LDOUBLE ptnLi_ = 0.0;
		for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) {

			LDOUBLE stateLi_ = 0.0;
			for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++)
				stateLi_ += //headLi_->liNdCube_[ptnNo_][0][headStateNo_] *
				    proChangeMat_[headStateNo_][tailStateNo_] *
				    tailLi_->liNdCube_[ptnNo_][0][tailStateNo_];

			ptnLi_ += stateLi_ * //stateFrqArr_[headStateNo_] *
			          headLi_->liNdCube_[ptnNo_][0][headStateNo_];
		}

		if (ptnLi_ < -ZERO)
			Utl::announceError (BAD_CMP_LI);

		double logPtnLi_ = cmpLogLi (ptnLi_);
		inLogLi_ += logPtnLi_ * ptnlist.weightArr_[ptnNo_];
	}

	// reduce by sum of all LogLikelihood to every process
	MPI_Allreduce(&inLogLi_, &global_inLogLi, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

	return global_inLogLi;
}
#endif
*/
/**********************************************************
*
*    SITE_SPECIFIC + INTERNAL branch
*
**********************************************************/

DMat20 probChangePtnMat_, prob_ptn_derv1, prob_ptn_derv2;

	/**
		compute the log likelihood of this tree 
		using site-specific rate
		via this an internal branch
		@return tree log-likelihood
	*/
double LiBr::cmpInLogLiSpecificRate () {
	opt_urtree.nCmpInLogLi_ ++;

	LiNd *headLi_;
	LiNd *tailLi_;

	headLi_ = &getHeadLiNd ();
	tailLi_ = &getTailLiNd ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	int nPtn_ = ptnlist.getNPtn ();

	Vec<double> *ptnRateArr_;
	ptnRateArr_ = myrate.getPtnRate ();

	int nState_ = mymodel.getNState ();

	double inLogLi_ = 0.0;
	#pragma omp parallel for   \
	private(probChangePtnMat_) \
	reduction(+: inLogLi_)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {

		double ptnRate_ = (*ptnRateArr_)[ptnNo_];
		//DMat20 probChangePtnMat_;
		cmpProbChangeFrq (len_ * ptnRate_, probChangePtnMat_);
		LDOUBLE ptnLi_ = 0.0;
		int start_addr = ptnNo_ * nState_;

		for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) {

			LDOUBLE stateLi_ = 0.0;
			for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++)
				stateLi_ += //headLi_->liNdCube_[ptnNo_][0][headStateNo_] *
				    probChangePtnMat_[headStateNo_][tailStateNo_] *
				    tailLi_->liNdCube_[start_addr + tailStateNo_];

			ptnLi_ += stateLi_ * //stateFrqArr_[headStateNo_] *
			          headLi_->liNdCube_[start_addr + headStateNo_];
		}

		if (ptnLi_ <= 0.0) {
			ptnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double logPtnLi_ = cmpLogLi (ptnLi_);
		inLogLi_ += logPtnLi_ * ptnlist.weightArr_[ptnNo_];
	}

	inLogLi_ += headLi_->liscale + tailLi_->liscale;
	return inLogLi_;
}

	/**
		compute the log likelihood of this tree 
		using site-specific rate
		via this internal branch
		@param calc_logli true if you want to compute log-likelihood
		@param logli_derv1 (OUT) 1st derivative of log-likelihood
		@param logli_derv2 (OUT) 2nd derivative of log-likelihood
		@return log-likelihood if calc_logli==true
	*/
double LiBr::cmpInLogLiDerivativeSpecificRate (double &logli_derv1_ret, double
&logli_derv2_ret, bool calc_logli) {
	opt_urtree.nCmpInLogLi_ ++;

	LiNd *headLi_;
	LiNd *tailLi_;

	headLi_ = &getHeadLiNd ();
	tailLi_ = &getTailLiNd ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	int nPtn_ = ptnlist.getNPtn ();

	Vec<double> *ptnRateArr_;
	ptnRateArr_ = myrate.getPtnRate ();

	int nState_ = mymodel.getNState ();

	double logli = 0.0;
	double logli_derv1 = 0.0;
	double logli_derv2 = 0.0;
	#pragma omp parallel for  \
	private(probChangePtnMat_, prob_ptn_derv1, prob_ptn_derv2)  \
	reduction(+: logli, logli_derv1, logli_derv2)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {

		double ptnRate_ = (*ptnRateArr_)[ptnNo_];
		//mymodel.cmpProbChange (len_ * ptnRate_, probChangePtnMat_);
		cmpProbChangeDerivativesFrq (len_ * ptnRate_, probChangePtnMat_,
		                             prob_ptn_derv1, prob_ptn_derv2);
		double rate_sqr = ptnRate_ * ptnRate_;
		for (int i = 0; i < nState_; i++)
			for (int j = 0; j < nState_; j++) {
				prob_ptn_derv1[i][j] *= ptnRate_;
				prob_ptn_derv2[i][j] *= rate_sqr;
			}

		LDOUBLE ptnLi_ = 0.0;
		LDOUBLE ptnLi_1 = 0.0;
		LDOUBLE ptnLi_2 = 0.0;
		int start_addr = ptnNo_ * nState_;

		for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) {

			LDOUBLE stateLi_ = 0.0;
			LDOUBLE stateLi_1 = 0.0;
			LDOUBLE stateLi_2 = 0.0;
			for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++) {
				stateLi_ += probChangePtnMat_[headStateNo_][tailStateNo_] *
				            tailLi_->liNdCube_[start_addr + tailStateNo_];
				stateLi_1 += prob_ptn_derv1[headStateNo_][tailStateNo_] *
				             tailLi_->liNdCube_[start_addr + tailStateNo_];
				stateLi_2 += prob_ptn_derv2[headStateNo_][tailStateNo_] *
				             tailLi_->liNdCube_[start_addr + tailStateNo_];
			}

			LDOUBLE temp = //stateFrqArr_[headStateNo_] *
			    headLi_->liNdCube_[start_addr + headStateNo_];
			ptnLi_ += stateLi_ * temp;
			ptnLi_1 += stateLi_1 * temp;
			ptnLi_2 += stateLi_2 * temp;
		}

		if (ptnLi_ <= 0.0) {
			ptnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double temp1 = ptnLi_1 / ptnLi_;
		double temp2 = ptnLi_2 / ptnLi_;
		logli_derv1 += temp1 * ptnlist.weightArr_[ptnNo_];
		logli_derv2 += (temp2 - temp1 * temp1) * ptnlist.weightArr_[ptnNo_];
		if (calc_logli)
			logli += cmpLogLi (ptnLi_) * ptnlist.weightArr_[ptnNo_];
	}
	logli_derv1_ret = logli_derv1;
	logli_derv2_ret = logli_derv2;
	logli += headLi_->liscale + tailLi_->liscale;
	return logli;
}

/**********************************************************
*
*    GAMMA + INTERNAL branch
*
**********************************************************/

DMat20 probChangeCube_[MAX_NUM_RATE];
DMat20 prob_cube_derv1[MAX_NUM_RATE];
DMat20 prob_cube_derv2[MAX_NUM_RATE];

	/**
		compute the log likelihood of this tree 
		using gamma rate
		via this an internal branch
		@return tree log-likelihood
	*/

double LiBr::cmpInLogLiGammaRate () {
	opt_urtree.nCmpInLogLi_ ++;

	LiNd *headLi_;
	LiNd *tailLi_;

	headLi_ = &getHeadLiNd ();
	tailLi_ = &getTailLiNd ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	int nState_ = mymodel.getNState ();
	int nRate_ = myrate.getNRate ();
	int rate_state = nRate_ * nState_;
	double probRate_ = myrate.getProb ();
	//double log_probRate = log(probRate_);

	int rateNo_;
	for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
		double rate_ = myrate.getRate (rateNo_);
		cmpProbChangeFrq (len_ * rate_, probChangeCube_[rateNo_]);
	}

	double inLogLi_ = 0.0;
	#pragma omp parallel for  private(rateNo_) reduction(+: inLogLi_)
	for (int ptnNo_ = 0; ptnNo_ < ptnlist.getNPtn (); ptnNo_ ++) {
		LDOUBLE ptnLi_ = 0.0;

		int start_addr = ptnNo_ * rate_state;

		for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
			LDOUBLE ratePtnLi_ = 0.0;
			int local_addr = start_addr + rateNo_ * nState_;

			for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) {

				LDOUBLE stateLi_ = 0.0;
				if (myrate.use_invar_site && rateNo_ == nRate_ - 1) {
					if (tailLi_->liNdCube_[local_addr + headStateNo_] > 0.0 &&
					        headLi_->liNdCube_[local_addr + headStateNo_] > 0.0)
						ratePtnLi_ += headLi_->liNdCube_[local_addr + headStateNo_] *
						              probChangeCube_[rateNo_][headStateNo_][headStateNo_] *
						              tailLi_->liNdCube_[local_addr + headStateNo_];
				} else {
					for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++)
						stateLi_ += //headLi_->liNdCube_[ptnNo_][rateNo_][headStateNo_] *
						    probChangeCube_[rateNo_][headStateNo_][tailStateNo_] *
						    tailLi_->liNdCube_[local_addr + tailStateNo_];

					ratePtnLi_ += stateLi_ * //stateFrqArr_[headStateNo_] *
					              headLi_->liNdCube_[local_addr + headStateNo_];
				}
			}

			if (!myrate.use_invar_site || rateNo_ < nRate_ - 1)
				ptnLi_ += ratePtnLi_  * probRate_;
			else
				ptnLi_ += ratePtnLi_ * myrate.prob_invar_site;
		}


		if (ptnLi_ <= 0.0) {
			ptnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double logPtnLi_ = cmpLogLi (ptnLi_) /* + log_probRate */;
		inLogLi_ += logPtnLi_ * ptnlist.weightArr_[ptnNo_];
	}

	inLogLi_ += headLi_->liscale + tailLi_->liscale;

	return inLogLi_;
}


	/**
		compute the log likelihood of this tree 
		using gamma rate
		via this internal branch
		@param calc_logli true if you want to compute log-likelihood
		@param logli_derv1 (OUT) 1st derivative of log-likelihood
		@param logli_derv2 (OUT) 2nd derivative of log-likelihood
		@return log-likelihood if calc_logli==true
	*/
double LiBr::cmpInLogLiDerivativeGammaRate (double &logli_derv1_ret, double
&logli_derv2_ret, bool calc_logli) {
	opt_urtree.nCmpInLogLi_ ++;

	LiNd *headLi_;
	LiNd *tailLi_;

	headLi_ = &getHeadLiNd ();
	tailLi_ = &getTailLiNd ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	int nState_ = mymodel.getNState ();
	int nRate_ = myrate.getNRate ();
	int rate_state = nRate_ * nState_;
	double probRate_ = myrate.getProb ();

	int rateNo_;
	for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
		double rate_ = myrate.getRate (rateNo_);
		//mymodel.cmpProbChange (len_ * rate_, probChangeCube_[rateNo_]);
		cmpProbChangeDerivativesFrq (len_ * rate_, probChangeCube_[rateNo_],
		                             prob_cube_derv1[rateNo_], prob_cube_derv2[rateNo_]);
		double rate_sqr = rate_ * rate_;
		for (int i = 0; i < nState_; i++)
			for (int j = 0; j < nState_; j++) {
				prob_cube_derv1[rateNo_][i][j] *= rate_;
				prob_cube_derv2[rateNo_][i][j] *= rate_sqr;
			}
	}

	double logli = 0.0;
	double logli_derv1 = 0.0;
	double logli_derv2 = 0.0;

	#pragma omp parallel for  private(rateNo_)  \
	reduction(+: logli, logli_derv1, logli_derv2)
	for (int ptnNo_ = 0; ptnNo_ < ptnlist.getNPtn (); ptnNo_ ++) {
		LDOUBLE ptnLi_ = 0.0;
		LDOUBLE ptnLi_1 = 0.0;
		LDOUBLE ptnLi_2 = 0.0;
		int start_addr = ptnNo_ * rate_state;

		for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
			LDOUBLE ratePtnLi_ = 0.0;
			LDOUBLE ratePtnLi_1 = 0.0;
			LDOUBLE ratePtnLi_2 = 0.0;
			bool invar_calc = (myrate.use_invar_site && rateNo_ == nRate_ - 1);
			int local_addr = start_addr + rateNo_ * nState_;

			for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) {

				LDOUBLE stateLi_ = 0.0;
				LDOUBLE stateLi_1 = 0.0;
				LDOUBLE stateLi_2 = 0.0;
				if (invar_calc)
					stateLi_ = probChangeCube_[rateNo_][headStateNo_][headStateNo_] *
					           tailLi_->liNdCube_[local_addr + headStateNo_];

				for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++) {
					if (!invar_calc)
						stateLi_ += probChangeCube_[rateNo_][headStateNo_][tailStateNo_] *
						            tailLi_->liNdCube_[local_addr + tailStateNo_];
					stateLi_1 += prob_cube_derv1[rateNo_][headStateNo_][tailStateNo_] *
					             tailLi_->liNdCube_[local_addr + tailStateNo_];
					stateLi_2 += prob_cube_derv2[rateNo_][headStateNo_][tailStateNo_] *
					             tailLi_->liNdCube_[local_addr + tailStateNo_];
				}

				LDOUBLE temp = //stateFrqArr_[headStateNo_] *
				    headLi_->liNdCube_[local_addr + headStateNo_];

				ratePtnLi_ += stateLi_ * temp;
				ratePtnLi_1 += stateLi_1 * temp;
				ratePtnLi_2 += stateLi_2 * temp;
			}

			if (!myrate.use_invar_site || rateNo_ < nRate_ - 1) {
				ptnLi_ += ratePtnLi_  * probRate_;
				ptnLi_1 += ratePtnLi_1 * probRate_;
				ptnLi_2 += ratePtnLi_2 * probRate_;
			} else {
				ptnLi_ += ratePtnLi_ * myrate.prob_invar_site;
				ptnLi_1 += ratePtnLi_1 * myrate.prob_invar_site;
				ptnLi_2 += ratePtnLi_2 * myrate.prob_invar_site;
			}
		}

		if (ptnLi_ <= 0.0) {
			ptnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double temp1 = ptnLi_1 / ptnLi_;
		double temp2 = ptnLi_2 / ptnLi_;
		logli_derv1 += temp1 * ptnlist.weightArr_[ptnNo_];
		logli_derv2 += (temp2 - temp1 * temp1) * ptnlist.weightArr_[ptnNo_];
		if (calc_logli)
			logli += cmpLogLi (ptnLi_) * ptnlist.weightArr_[ptnNo_];
	}
	logli_derv1_ret = logli_derv1;
	logli_derv2_ret = logli_derv2;
	logli += headLi_->liscale + tailLi_->liscale;
	return logli;
}


void print_li_info(int classNo_, PVec tail, PVec head, DMat20 *probChange) {
	int nState_ = mymodel.getNState ();
	//int nClass_ = myrate.getNRate ();
	//LDOUBLE ratePtnLi_ = 0.0;

	for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) {

		LDOUBLE stateLi_ = 0.0;
		cout << headStateNo_ << "  ";
		for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++) {
			stateLi_ += probChange[classNo_][headStateNo_][tailStateNo_] *
				tail[tailStateNo_];
			if (tail[tailStateNo_] < 0.0)
				cout << "{" << tail[tailStateNo_] << "}  ";
			else
				cout << tail[tailStateNo_] << "  ";
		}

		if (stateLi_ < 0.0)
			cout << "[" << stateLi_ << "] " << endl;
		else
			cout << stateLi_ << endl;

		stateLi_ *= head[headStateNo_];
	}

}


	//-----------------------------------------------
	//-  codon-based Nielsen Yang Model
	//-----------------------------------------------
	/**
		compute the log likelihood of this tree 
		for heterogenous omega model
		via this an internal branch
		@return tree log-likelihood
	*/
double LiBr::cmpInLogLiSelection () {
	opt_urtree.nCmpInLogLi_ ++;

	LiNd *headLi_;
	LiNd *tailLi_;

	headLi_ = &getHeadLiNd ();
	tailLi_ = &getTailLiNd ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	int nState_ = mymodel.getNState ();
	int nClass_ = myrate.getNRate ();
	int class_state = nClass_ * nState_;

	//double log_probRate = log(probRate_);

	int classNo_;
	for (classNo_ = 0; classNo_ < nClass_; classNo_ ++) {
		cmpProbChangeFrq (len_, classNo_, probChangeCube_[classNo_]);
		//mymodel.cmpProbChange (len_, classNo_, probChangeCube_[classNo_]);
	}

	double inLogLi_ = 0.0;
	#pragma omp parallel for  private(classNo_)  \
	reduction(+: inLogLi_)
	for (int ptnNo_ = 0; ptnNo_ < ptnlist.getNPtn (); ptnNo_ ++) {
		LDOUBLE ptnLi_ = 0.0;
		int start_addr = ptnNo_ * class_state;

		for (classNo_ = 0; classNo_ < nClass_; classNo_ ++) {
			LDOUBLE ratePtnLi_ = 0.0;
			int local_addr = start_addr + classNo_ * nState_;

			for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) {

				LDOUBLE stateLi_ = 0.0;
				for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++)
					stateLi_ += //headLi_->liNdCube_[ptnNo_][classNo_][headStateNo_] *
					    probChangeCube_[classNo_][headStateNo_][tailStateNo_] *
					    tailLi_->liNdCube_[local_addr + tailStateNo_];

				ratePtnLi_ += stateLi_ * //mymodel.stateFrqArr_[headStateNo_] *
				              headLi_->liNdCube_[local_addr + headStateNo_];
			}

			ptnLi_ += ratePtnLi_ * mymodel.getClassProb(classNo_);
#ifdef PRINT_NEGATIVE_INFO
			if (ratePtnLi_ * mymodel.getClassProb(classNo_) < 0.0) 
				print_li_info(classNo_, tailLi_->liNdCube_ + local_addr, 
					headLi_->liNdCube_ + local_addr, probChangeCube_);
#endif
		}


		if (ptnLi_ <= 0.0) {
			ptnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double logPtnLi_ = cmpLogLi (ptnLi_) /* + log_probRate */;
		inLogLi_ += logPtnLi_ * ptnlist.weightArr_[ptnNo_];
	}

	inLogLi_ += headLi_->liscale + tailLi_->liscale;
	return inLogLi_;
}

	/**
		compute the log likelihood of this tree 
		for heterogenous omega model
		via this internal branch
		@param calc_logli true if you want to compute log-likelihood
		@param logli_derv1 (OUT) 1st derivative of log-likelihood
		@param logli_derv2 (OUT) 2nd derivative of log-likelihood
		@return log-likelihood if calc_logli==true
	*/

double LiBr::cmpInLogLiDerivativeSelection (double &logli_derv1_ret, double
&logli_derv2_ret, bool calc_logli) {
	opt_urtree.nCmpInLogLi_ ++;

	LiNd *headLi_;
	LiNd *tailLi_;

	headLi_ = &getHeadLiNd ();
	tailLi_ = &getTailLiNd ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	int nState_ = mymodel.getNState ();
	int nRate_ = myrate.getNRate ();
	int class_state = nRate_ * nState_;

	int rateNo_;
	for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
		//mymodel.cmpProbChange (len_ * rate_, probChangeCube_[rateNo_]);
		cmpProbChangeDerivativesFrq (len_, rateNo_, probChangeCube_[rateNo_],
		                             prob_cube_derv1[rateNo_], prob_cube_derv2[rateNo_]);
		/*
		double rate_sqr = rate_ * rate_;
		for (int i = 0; i < nState_; i++)
			for (int j = 0; j < nState_; j++) {
				prob_cube_derv1[rateNo_][i][j] *= rate_;
				prob_cube_derv2[rateNo_][i][j] *= rate_sqr;
			}
		*/
	}

	double logli = 0.0;
	double logli_derv1 = 0.0;
	double logli_derv2 = 0.0;

	#pragma omp parallel for  private(rateNo_)  \
	reduction(+: logli, logli_derv1, logli_derv2)
	for (int ptnNo_ = 0; ptnNo_ < ptnlist.getNPtn (); ptnNo_ ++) {
		LDOUBLE ptnLi_ = 0.0;
		LDOUBLE ptnLi_1 = 0.0;
		LDOUBLE ptnLi_2 = 0.0;
		int start_addr = ptnNo_ * class_state;

		for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
			LDOUBLE ratePtnLi_ = 0.0;
			LDOUBLE ratePtnLi_1 = 0.0;
			LDOUBLE ratePtnLi_2 = 0.0;
			int local_addr = start_addr + rateNo_ * nState_;

			for (int headStateNo_ = 0; headStateNo_ < nState_; headStateNo_ ++) {

				LDOUBLE stateLi_ = 0.0;
				LDOUBLE stateLi_1 = 0.0;
				LDOUBLE stateLi_2 = 0.0;

				for (int tailStateNo_ = 0; tailStateNo_ < nState_; tailStateNo_ ++) {
					stateLi_ += probChangeCube_[rateNo_][headStateNo_][tailStateNo_] *
					            tailLi_->liNdCube_[local_addr + tailStateNo_];
					stateLi_1 += prob_cube_derv1[rateNo_][headStateNo_][tailStateNo_] *
					             tailLi_->liNdCube_[local_addr + tailStateNo_];
					stateLi_2 += prob_cube_derv2[rateNo_][headStateNo_][tailStateNo_] *
					             tailLi_->liNdCube_[local_addr + tailStateNo_];
				}

				LDOUBLE temp = //stateFrqArr_[headStateNo_] *
				    headLi_->liNdCube_[local_addr + headStateNo_];

				ratePtnLi_ += stateLi_ * temp;
				ratePtnLi_1 += stateLi_1 * temp;
				ratePtnLi_2 += stateLi_2 * temp;
			}

			ptnLi_ += ratePtnLi_  * mymodel.getClassProb(rateNo_);
			ptnLi_1 += ratePtnLi_1 * mymodel.getClassProb(rateNo_);
			ptnLi_2 += ratePtnLi_2 * mymodel.getClassProb(rateNo_);
#ifdef PRINT_NEGATIVE_INFO
			if (ratePtnLi_ * mymodel.getClassProb(rateNo_) < 0.0) 
				print_li_info(rateNo_, tailLi_->liNdCube_ + local_addr, headLi_->liNdCube_ + local_addr, probChangeCube_);
#endif
		}

		if (ptnLi_ <= 0.0) {
			ptnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double temp1 = ptnLi_1 / ptnLi_;
		double temp2 = ptnLi_2 / ptnLi_;
		logli_derv1 += temp1 * ptnlist.weightArr_[ptnNo_];
		logli_derv2 += (temp2 - temp1 * temp1) * ptnlist.weightArr_[ptnNo_];
		if (calc_logli)
			logli += cmpLogLi (ptnLi_) * ptnlist.weightArr_[ptnNo_];
	}
	logli_derv1_ret = logli_derv1;
	logli_derv2_ret = logli_derv2;
	logli += headLi_->liscale + tailLi_->liscale;
	return logli;
}



/**********************************************************
*
*    UNIFORM + EXTERNAL branch
*
**********************************************************/

//-------------------------------------------------------------------
	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
	*/

double LiBr::cmpExLogLiUniformRate () {

	Seq *seq_;
	seq_ = &alignment.getSeq (smaNdNo_);

	int nState_ = mymodel.getNState ();

	cmpProbChangeFrq (len_, proChangeMat_);

	double logLi_ = 0.0;
	int nPtn_ = ptnlist.getNPtn ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	#pragma omp parallel for   \
	reduction(+: logLi_)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
		int siteNo_ = ptnlist.siteArr_ [ptnNo_];
		int exStateNo_ = (*seq_).items_[siteNo_];
		LDOUBLE exPtnLi_ = 0.0;
		int start_addr = ptnNo_ * nState_;

		if (exStateNo_ < nState_) {
			for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++)
				exPtnLi_ += greLiNd_->liNdCube_[start_addr + inStateNo_] *
				            proChangeMat_[exStateNo_][inStateNo_];

			//exPtnLi_ *= stateFrqArr_[exStateNo_];
		} else {
			for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {

				double exStateLi_ = 0.0;
				if (exStateNo_ == BS_UNKNOWN)
					exStateLi_ = mymodel.stateFrqArr_[inStateNo_];
				else
					for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++)
						exStateLi_ += proChangeMat_[inStateNo_][exStateNo_] *
						              smaLiNd_->liNdCube_[start_addr + exStateNo_];

				exPtnLi_ += greLiNd_->liNdCube_[start_addr + inStateNo_] * exStateLi_;
				//* stateFrqArr_[inStateNo_] ;
			}
		}

		if (exPtnLi_ <= 0.0) {
			exPtnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double logExPtnLi_ = cmpLogLi (exPtnLi_);
		//logExPtnLi_ += greLiNd_->liscale[0][ptnNo_] + smaLiNd_->liscale[0][ptnNo_];
		logLi_ += logExPtnLi_ * ptnlist.weightArr_[ptnNo_];
		//std::cout << ptnNo_ << " / " << logLi_ << endl;
	}

	logLi_ += greLiNd_->liscale + smaLiNd_->liscale;
	return logLi_;
}

/*
#ifdef PARALLEL
//-------------------------------------------------------------------
//parallel compute the log likelihood of this tree via this branch
double LiBr::par_cmpExLogLiUniformRate () {
	Cube *greLiNd_->liNdCube_;
	greLiNd_->liNdCube_ = &greLiNd_->liNdCube_;

	Cube *smaLiNd_->liNdCube_;
	smaLiNd_->liNdCube_ = &smaLiNd_->liNdCube_;

	Seq *seq_;
	seq_ = &alignment.getSeq (smaNdNo_);

	int nState_ = mymodel.getNState ();

	cmpProbChangeFrq (len_, proChangeMat_);


	double logLi_ = 0.0;
	double global_logLi;
	int nPtn_ = ptnlist.getNPtn ();

	int work_load = (int) ceil((double)nPtn_ / mpi_size);
	int work_start = work_load * mpi_myrank;
	int work_end = work_start + work_load;
	if (work_end > nPtn_) work_end = nPtn_;


	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	for (int ptnNo_ = work_start; ptnNo_ < work_end ; ptnNo_ ++) {
		int siteNo_ = ptnlist.siteArr_ [ptnNo_];
		int exStateNo_ = (*seq_).items_[siteNo_];
		LDOUBLE exPtnLi_ = 0.0;

		if (exStateNo_ < nState_) {
			for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++)
				exPtnLi_ += greLiNd_->liNdCube_[ptnNo_][0][inStateNo_] *
				            proChangeMat_[exStateNo_][inStateNo_];

			//exPtnLi_ *= stateFrqArr_[exStateNo_];
		} else {
			for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {

				double exStateLi_ = 0.0;
				if (exStateNo_ == BS_UNKNOWN)
					exStateLi_ = mymodel.stateFrqArr_[inStateNo_];
				else
					for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++)
						exStateLi_ += proChangeMat_[inStateNo_][exStateNo_] *
						              smaLiNd_->liNdCube_[ptnNo_][0][exStateNo_];

				exPtnLi_ += greLiNd_->liNdCube_[ptnNo_][0][inStateNo_]  * exStateLi_;
				// stateFrqArr_[inStateNo_];
			}
		}

		if (exPtnLi_ < -ZERO) {
			Utl::announceError (BAD_CMP_LI);
		}

		double logExPtnLi_ = cmpLogLi (exPtnLi_);
		logLi_ += logExPtnLi_ * ptnlist.weightArr_[ptnNo_];
		//std::cout << ptnNo_ << " / " << logLi_ << endl;
	}

	// take the sum of all process log likelihood, put result into global_logli
	MPI_Allreduce(&logLi_, &global_logLi, 1, MPI_DOUBLE, MPI_SUM, MPI_COMM_WORLD);

	return global_logLi;
}
#endif // PARALLEL
*/

	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
	*/
double LiBr::cmpExLogLiDerivativeUniformRate (double &logli_derv1_ret, double
&logli_derv2_ret, bool calc_logli) {

	Seq *seq_;
	seq_ = &alignment.getSeq (smaNdNo_);

	int nState_ = mymodel.getNState ();

	//mymodel.cmpProbChange (len_, proChangeMat_);

	cmpProbChangeDerivativesFrq (len_, proChangeMat_, derv1Mat, derv2Mat);

	double logli = 0.0;
	double logli_derv1 = 0.0;
	double logli_derv2 = 0.0;

	int nPtn_ = ptnlist.getNPtn ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	#pragma omp parallel for   \
	reduction(+: logli, logli_derv1, logli_derv2)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
		int siteNo_ = ptnlist.siteArr_ [ptnNo_];
		int exStateNo_ = (*seq_).items_[siteNo_];
		LDOUBLE exPtnLi_ = 0.0;
		LDOUBLE exPtnLi_1 = 0.0;
		LDOUBLE exPtnLi_2 = 0.0;
		int start_addr = ptnNo_ * nState_;

		if (exStateNo_ < nState_) {
			for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
				exPtnLi_ += greLiNd_->liNdCube_[start_addr + inStateNo_] *
				            proChangeMat_[exStateNo_][inStateNo_];
				exPtnLi_1 += greLiNd_->liNdCube_[start_addr + inStateNo_] *
				             derv1Mat[exStateNo_][inStateNo_];
				exPtnLi_2 += greLiNd_->liNdCube_[start_addr + inStateNo_] *
				             derv2Mat[exStateNo_][inStateNo_];
			}

			//exPtnLi_ *= stateFrqArr_[exStateNo_];
			//exPtnLi_1 *= stateFrqArr_[exStateNo_];
			//exPtnLi_2 *= stateFrqArr_[exStateNo_];
		} else {
			for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {

				double exStateLi_ = 0.0;
				double exStateLi_1 = 0.0;
				double exStateLi_2 = 0.0;
				if (exStateNo_ == BS_UNKNOWN)
					exStateLi_ = mymodel.stateFrqArr_[inStateNo_];
				else
					for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++) {
						exStateLi_ += proChangeMat_[inStateNo_][exStateNo_] *
						              smaLiNd_->liNdCube_[start_addr + exStateNo_];
						exStateLi_1 += derv1Mat[inStateNo_][exStateNo_] *
						               smaLiNd_->liNdCube_[start_addr + exStateNo_];
						exStateLi_2 += derv2Mat[inStateNo_][exStateNo_] *
						               smaLiNd_->liNdCube_[start_addr + exStateNo_];
					}

				LDOUBLE temp = //stateFrqArr_[inStateNo_] *
				    greLiNd_->liNdCube_[start_addr + inStateNo_];
				exPtnLi_ += exStateLi_ * temp;
				exPtnLi_1 += exStateLi_1 * temp;
				exPtnLi_2 += exStateLi_2 * temp;
			}
		}

		if (exPtnLi_ <= 0.0) {
			exPtnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double temp1 = exPtnLi_1 / exPtnLi_;
		double temp2 = exPtnLi_2 / exPtnLi_;
		if (isnan(temp1) || isnan(temp2)) {
			Utl::announceError("NaN in cmpExLogLiDerivativeUniformRate");
		}
		logli_derv1 += temp1 * ptnlist.weightArr_[ptnNo_];
		logli_derv2 += (temp2 - temp1 * temp1) * ptnlist.weightArr_[ptnNo_];
		if (calc_logli) {
			double logExPtnLi_ = cmpLogLi (exPtnLi_);
			logli += logExPtnLi_ * ptnlist.weightArr_[ptnNo_];
		}
	}
	logli_derv1_ret = logli_derv1;
	logli_derv2_ret = logli_derv2;

	logli += greLiNd_->liscale + smaLiNd_->liscale;
	return logli;
}

/**********************************************************
*
*    SITE_SPECIFIC + EXTERNAL branch
*
**********************************************************/

DVec20 probChangePtnVec_, prob_vec_derv1, prob_vec_derv2;

	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
		@return TODO
	*/
double LiBr::cmpExLogLiSpecificRate () {

	Seq *seq_;
	seq_ = &alignment.getSeq (smaNdNo_);

	int nState_ = mymodel.getNState ();

	int nPtn_ = ptnlist.getNPtn ();

	Vec<double> *ptnRateArr_;
	ptnRateArr_ = myrate.getPtnRate ();

	double logLi_ = 0.0;

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);


	#pragma omp parallel for   \
	private(probChangePtnVec_, probChangePtnMat_)  \
	reduction(+: logLi_)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {

		int siteNo_ = ptnlist.siteArr_ [ptnNo_];
		int exStateNo_ = (*seq_).items_[siteNo_];
		LDOUBLE exPtnLi_ = 0.0;
		int start_addr = ptnNo_ * nState_;

		if (exStateNo_ < nState_) {

			double ptnRate_ = (*ptnRateArr_)[ptnNo_];
			mymodel.cmpProbChange (exStateNo_, len_ * ptnRate_, probChangePtnVec_);

			for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
				exPtnLi_ += greLiNd_->liNdCube_[start_addr + inStateNo_] *
				            probChangePtnVec_[inStateNo_];
			}

			exPtnLi_ *= mymodel.stateFrqArr_[exStateNo_];
		} else {
			double ptnRate_ = (*ptnRateArr_)[ptnNo_];
			//DMat20 probChangePtnMat_;
			cmpProbChangeFrq (len_ * ptnRate_, probChangePtnMat_);

			for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {

				double exStateLi_ = 0.0;
				if (exStateNo_ == BS_UNKNOWN)
					exStateLi_ = mymodel.stateFrqArr_[inStateNo_];
				else
					for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++)
						exStateLi_ += probChangePtnMat_[inStateNo_][exStateNo_] *
						              smaLiNd_->liNdCube_[start_addr + exStateNo_];

				exPtnLi_ += greLiNd_->liNdCube_[start_addr + inStateNo_] * exStateLi_;
				//stateFrqArr_[inStateNo_] *
			}
		}

		if (exPtnLi_ <= 0.0) {
			exPtnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double logExPtnLi_ = cmpLogLi (exPtnLi_);
		logLi_ += logExPtnLi_ * ptnlist.weightArr_[ptnNo_];
		//std::cout << ptnNo_ << " / " << logLi_ << endl;
	}

	logLi_ += greLiNd_->liscale + smaLiNd_->liscale;
	return logLi_;
}

	/**
		compute the log likelihood of this tree via this an external branch in case of one rate
		@return TODO
	*/
double LiBr::cmpExLogLiDerivativeSpecificRate (double &logli_derv1_ret, double
&logli_derv2_ret, bool calc_logli) {

	Seq *seq_;
	seq_ = &alignment.getSeq (smaNdNo_);

	int nState_ = mymodel.getNState ();

	int nPtn_ = ptnlist.getNPtn ();

	Vec<double> *ptnRateArr_;
	ptnRateArr_ = myrate.getPtnRate ();

	double logli = 0.0;
	double logli_derv1 = 0.0;
	double logli_derv2 = 0.0;

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);


	#pragma omp parallel for   \
	private(probChangePtnVec_, prob_vec_derv1, prob_vec_derv2, \
		probChangePtnMat_, prob_ptn_derv1, prob_ptn_derv2)  \
	reduction(+: logli, logli_derv1, logli_derv2)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {

		int siteNo_ = ptnlist.siteArr_ [ptnNo_];
		int exStateNo_ = (*seq_).items_[siteNo_];
		LDOUBLE exPtnLi_ = 0.0;
		LDOUBLE exPtnLi_1 = 0.0;
		LDOUBLE exPtnLi_2 = 0.0;
		int start_addr = ptnNo_ * nState_;

		if (exStateNo_ < nState_) {

			double ptnRate_ = (*ptnRateArr_)[ptnNo_];
			DVec20 probChangePtnVec_;
			//mymodel.cmpProbChange (exStateNo_, len_ * ptnRate_, probChangePtnVec_);
			mymodel.cmpProbChangeDerivatives (exStateNo_, len_ * ptnRate_,
			                                  probChangePtnVec_, prob_vec_derv1, prob_vec_derv2);
			double rate_sqr = ptnRate_ * ptnRate_;
			for (int j = 0; j < nState_; j++) {
				prob_vec_derv1[j] *= ptnRate_;
				prob_vec_derv2[j] *= rate_sqr;
			}


			for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
				exPtnLi_ += greLiNd_->liNdCube_[start_addr + inStateNo_] *
				            probChangePtnVec_[inStateNo_];
				exPtnLi_1 += greLiNd_->liNdCube_[start_addr + inStateNo_] *
				             prob_vec_derv1[inStateNo_];
				exPtnLi_2 += greLiNd_->liNdCube_[start_addr + inStateNo_] *
				             prob_vec_derv2[inStateNo_];
			}

			exPtnLi_ *= mymodel.stateFrqArr_[exStateNo_];
			exPtnLi_1 *= mymodel.stateFrqArr_[exStateNo_];
			exPtnLi_2 *= mymodel.stateFrqArr_[exStateNo_];

		} else {
			double ptnRate_ = (*ptnRateArr_)[ptnNo_];
			//DMat20 probChangePtnMat_;
			//mymodel.cmpProbChange (len_ * ptnRate_, probChangePtnMat_);
			cmpProbChangeDerivativesFrq (len_ * ptnRate_, probChangePtnMat_,
			                             prob_ptn_derv1, prob_ptn_derv2);
			double rate_sqr = ptnRate_ * ptnRate_;
			for (int i = 0; i < nState_; i++)
				for (int j = 0; j < nState_; j++) {
					prob_ptn_derv1[i][j] *= ptnRate_;
					prob_ptn_derv2[i][j] *= rate_sqr;
				}

			for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {

				double exStateLi_ = 0.0;
				double exStateLi_1 = 0.0;
				double exStateLi_2 = 0.0;
				if (exStateNo_ == BS_UNKNOWN)
					exStateLi_ = mymodel.stateFrqArr_[inStateNo_];
				else
					for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++) {
						exStateLi_ += probChangePtnMat_[inStateNo_][exStateNo_] *
						              smaLiNd_->liNdCube_[start_addr + exStateNo_];
						exStateLi_1 += prob_ptn_derv1[inStateNo_][exStateNo_] *
						               smaLiNd_->liNdCube_[start_addr + exStateNo_];
						exStateLi_2 += prob_ptn_derv2[inStateNo_][exStateNo_] *
						               smaLiNd_->liNdCube_[start_addr + exStateNo_];
					}

				LDOUBLE temp = greLiNd_->liNdCube_[start_addr + inStateNo_];
				// * stateFrqArr_[inStateNo_];
				exPtnLi_ += temp * exStateLi_;
				exPtnLi_1 += temp * exStateLi_1;
				exPtnLi_2 += temp * exStateLi_2;
			}
		}

		if (exPtnLi_ <= 0.0) {
			exPtnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double temp1 = exPtnLi_1 / exPtnLi_;
		double temp2 = exPtnLi_2 / exPtnLi_;
		logli_derv1 += temp1 * ptnlist.weightArr_[ptnNo_];
		logli_derv2 += (temp2 - temp1 * temp1) * ptnlist.weightArr_[ptnNo_];
		if (calc_logli)
			logli += cmpLogLi (exPtnLi_) * ptnlist.weightArr_[ptnNo_];

	}
	
	logli_derv1_ret = logli_derv1;
	logli_derv2_ret = logli_derv2;
	logli += greLiNd_->liscale + smaLiNd_->liscale;
	
	return logli;
}



/**********************************************************
*
*    GAMMA + EXTERNAL branch
*
**********************************************************/
//-------------------------------------------------------------------
//compute the log likelihood of this tree via this branch
double LiBr::cmpExLogLiGammaRate () {

	Seq *seq_;
	seq_ = &alignment.getSeq (smaNdNo_);

	int nState_ = mymodel.getNState ();
	int nRate_ = myrate.getNRate ();
	int rate_state = nRate_ * nState_;
	double probRate_ = myrate.getProb ();

	//DMat20 probChangeCube_[MAX_NUM_RATE];
	int rateNo_;
	for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
		double rate_ = myrate.getRate (rateNo_);
		cmpProbChangeFrq (len_ * rate_, probChangeCube_[rateNo_]);
	}

	double logLi_ = 0.0;
	int nPtn_ = ptnlist.getNPtn ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	#pragma omp parallel for  private(rateNo_) reduction(+: logLi_)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
		int siteNo_ = ptnlist.siteArr_ [ptnNo_];
		int exStateNo_ = (*seq_).items_[siteNo_];
		LDOUBLE exPtnLi_ = 0.0;
		int start_addr = ptnNo_ * rate_state;

		for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
			int local_addr = start_addr + rateNo_ * nState_;

			LDOUBLE rateExPtnLi_ = 0.0;
			if (exStateNo_ < nState_) {
				if (myrate.use_invar_site && rateNo_ == nRate_ - 1)
					rateExPtnLi_ = greLiNd_->liNdCube_[local_addr + exStateNo_] *
					               probChangeCube_[rateNo_][exStateNo_][exStateNo_];
				else
					for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++)
						rateExPtnLi_ += greLiNd_->liNdCube_[local_addr + inStateNo_] *
						                probChangeCube_[rateNo_][exStateNo_][inStateNo_];

				//rateExPtnLi_ *= stateFrqArr_[exStateNo_];
			} else {
				for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
					double exStateLi_ = 0.0;

					if (exStateNo_ == BS_UNKNOWN)
						exStateLi_ = mymodel.stateFrqArr_[inStateNo_];
					else {
						if (myrate.use_invar_site && rateNo_ == nRate_ - 1)
							exStateLi_ = probChangeCube_[rateNo_][inStateNo_][inStateNo_] *
							             smaLiNd_->liNdCube_[local_addr + inStateNo_];
						else
							for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++)
								exStateLi_ += probChangeCube_[rateNo_][inStateNo_][exStateNo_] *
								              smaLiNd_->liNdCube_[local_addr + exStateNo_];
					}

					rateExPtnLi_ += greLiNd_->liNdCube_[local_addr + inStateNo_] * exStateLi_;
					//stateFrqArr_[inStateNo_] *
				}
			}

			if (!myrate.use_invar_site || rateNo_ < nRate_ - 1)
				exPtnLi_ += rateExPtnLi_ * probRate_;
			else
				exPtnLi_ += rateExPtnLi_ * myrate.prob_invar_site;
		}

		if (exPtnLi_ <= 0.0) {
			exPtnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double logExPtnLi_ = cmpLogLi (exPtnLi_);
		logLi_ += logExPtnLi_ * ptnlist.weightArr_[ptnNo_];
		//std::cout << ptnNo_ << " / " << logLi_ << endl;
	}

	logLi_ += greLiNd_->liscale + smaLiNd_->liscale;
	return logLi_;

}

//-------------------------------------------------------------------
//compute the log likelihood of this tree via this branch
double LiBr::cmpExLogLiDerivativeGammaRate (double &logli_derv1_ret, double
&logli_derv2_ret, bool calc_logli) {

	Seq *seq_;
	seq_ = &alignment.getSeq (smaNdNo_);

	int nState_ = mymodel.getNState ();
	int nRate_ = myrate.getNRate ();
	int rate_state = nRate_ * nState_;
	double probRate_ = myrate.getProb ();

	int rateNo_;
	for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
		double rate_ = myrate.getRate (rateNo_);
		//mymodel.cmpProbChange (len_ * rate_, probChangeCube_[rateNo_]);
		cmpProbChangeDerivativesFrq (len_ * rate_, probChangeCube_[rateNo_],
		                             prob_cube_derv1[rateNo_], prob_cube_derv2[rateNo_]);
		double rate_sqr = rate_ * rate_;
		for (int i = 0; i < nState_; i++)
			for (int j = 0; j < nState_; j++) {
				prob_cube_derv1[rateNo_][i][j] *= rate_;
				prob_cube_derv2[rateNo_][i][j] *= rate_sqr;
			}
	}

	double logli = 0.0;
	double logli_derv1 = 0.0;
	double logli_derv2 = 0.0;
	int nPtn_ = ptnlist.getNPtn ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	#pragma omp parallel for  private(rateNo_)  \
	reduction(+: logli, logli_derv1, logli_derv2)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
		int siteNo_ = ptnlist.siteArr_ [ptnNo_];
		int exStateNo_ = (*seq_).items_[siteNo_];
		LDOUBLE exPtnLi_ = 0.0;
		LDOUBLE exPtnLi_1 = 0.0;
		LDOUBLE exPtnLi_2 = 0.0;
		int start_addr = ptnNo_ * rate_state;

		for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
			int local_addr = start_addr + rateNo_ * nState_;

			LDOUBLE rateExPtnLi_ = 0.0;
			LDOUBLE rateExPtnLi_1 = 0.0;
			LDOUBLE rateExPtnLi_2 = 0.0;
			bool invar_calc = (myrate.use_invar_site && rateNo_ == nRate_ - 1);
			if (exStateNo_ < nState_) {
				if (invar_calc)
					rateExPtnLi_ = greLiNd_->liNdCube_[local_addr + exStateNo_] *
					               probChangeCube_[rateNo_][exStateNo_][exStateNo_];
				for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
					if (!invar_calc)
						rateExPtnLi_ += greLiNd_->liNdCube_[local_addr + inStateNo_] *
						                probChangeCube_[rateNo_][exStateNo_][inStateNo_];
					rateExPtnLi_1 += greLiNd_->liNdCube_[local_addr + inStateNo_] *
					                 prob_cube_derv1[rateNo_][exStateNo_][inStateNo_];
					rateExPtnLi_2 += greLiNd_->liNdCube_[local_addr + inStateNo_] *
					                 prob_cube_derv2[rateNo_][exStateNo_][inStateNo_];
				}

				//rateExPtnLi_ *= stateFrqArr_[exStateNo_];
				//rateExPtnLi_1 *= stateFrqArr_[exStateNo_];
				//rateExPtnLi_2 *= stateFrqArr_[exStateNo_];
			} else {
				for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
					double exStateLi_ = 0.0;
					double exStateLi_1 = 0.0;
					double exStateLi_2 = 0.0;

					if (exStateNo_ == BS_UNKNOWN)
						exStateLi_ = mymodel.stateFrqArr_[inStateNo_];
					else {
						if (invar_calc)
							exStateLi_ = probChangeCube_[rateNo_][inStateNo_][inStateNo_] *
							             smaLiNd_->liNdCube_[local_addr + inStateNo_];

						for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++) {
							if (!invar_calc)
								exStateLi_ += probChangeCube_[rateNo_][inStateNo_][exStateNo_] *
								              smaLiNd_->liNdCube_[local_addr + exStateNo_];
							exStateLi_1 += prob_cube_derv1[rateNo_][inStateNo_][exStateNo_] *
							               smaLiNd_->liNdCube_[local_addr + exStateNo_];
							exStateLi_2 += prob_cube_derv2[rateNo_][inStateNo_][exStateNo_] *
							               smaLiNd_->liNdCube_[local_addr + exStateNo_];
						}
					}

					LDOUBLE temp = greLiNd_->liNdCube_[local_addr + inStateNo_];
					//* stateFrqArr_[inStateNo_];
					rateExPtnLi_ += temp * exStateLi_;
					rateExPtnLi_1 += temp * exStateLi_1;
					rateExPtnLi_2 += temp * exStateLi_2;
				}
			}

			if (!myrate.use_invar_site || rateNo_ < nRate_ - 1) {
				exPtnLi_ += rateExPtnLi_ * probRate_;
				exPtnLi_1 += rateExPtnLi_1 * probRate_;
				exPtnLi_2 += rateExPtnLi_2 * probRate_;
			} else {
				exPtnLi_ += rateExPtnLi_ * myrate.prob_invar_site;
				exPtnLi_1 += rateExPtnLi_1 * myrate.prob_invar_site;
				exPtnLi_2 += rateExPtnLi_2 * myrate.prob_invar_site;
			}
		}

		if (exPtnLi_ <= 0.0) {
			exPtnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double temp1 = exPtnLi_1 / exPtnLi_;
		double temp2 = exPtnLi_2 / exPtnLi_;
		logli_derv1 += temp1 * ptnlist.weightArr_[ptnNo_];
		logli_derv2 += (temp2 - temp1 * temp1) * ptnlist.weightArr_[ptnNo_];
		if (calc_logli)
			logli += cmpLogLi (exPtnLi_) * ptnlist.weightArr_[ptnNo_];

	}
	logli_derv1_ret = logli_derv1;
	logli_derv2_ret = logli_derv2;
	logli += greLiNd_->liscale + smaLiNd_->liscale;
	
	return logli;
}

void LiBr::empiricalBayesGammaRate (ostream &out) {

	Seq *seq_;
	seq_ = &alignment.getSeq (smaNdNo_);

	int nState_ = mymodel.getNState ();
	int nRate_ = myrate.getNRate ();
	int rate_state = nRate_ * nState_;
	double probRate_ = myrate.getProb ();
	int ptnNo_;

	//DMat20 probChangeCube_[MAX_NUM_RATE];
	int rateNo_;
	/*
	for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
		double rate_ = myrate.getRate (rateNo_);
		cout << rate_ << "  ";
		cmpProbChangeFrq (len_ * rate_, probChangeCube_[rateNo_]);
	}
	cout << endl;
	*/
	int nPtn_ = ptnlist.getNPtn ();
	Vec<double> *ptn_rate = myrate.getPtnRate();
	ptn_rate->set(nPtn_, nPtn_);

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	#pragma omp parallel for  private(rateNo_) reduction(+: logLi_)
	for (ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
		int siteNo_ = ptnlist.siteArr_ [ptnNo_];
		int exStateNo_ = (*seq_).items_[siteNo_];
		LDOUBLE exPtnLi_ = 0.0;
		int start_addr = ptnNo_ * rate_state;
		(*ptn_rate)[ptnNo_] = 0.0;

		for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
			int local_addr = start_addr + rateNo_ * nState_;

			LDOUBLE rateExPtnLi_ = 0.0;
			if (exStateNo_ < nState_) {
				if (myrate.use_invar_site && rateNo_ == nRate_ - 1)
					rateExPtnLi_ = greLiNd_->liNdCube_[local_addr + exStateNo_] *
					               probChangeCube_[rateNo_][exStateNo_][exStateNo_];
				else
					for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++)
						rateExPtnLi_ += greLiNd_->liNdCube_[local_addr + inStateNo_] *
						                probChangeCube_[rateNo_][exStateNo_][inStateNo_];

				//rateExPtnLi_ *= stateFrqArr_[exStateNo_];
			} else {
				for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
					double exStateLi_ = 0.0;

					if (exStateNo_ == BS_UNKNOWN)
						exStateLi_ = mymodel.stateFrqArr_[inStateNo_];
					else {
						if (myrate.use_invar_site && rateNo_ == nRate_ - 1)
							exStateLi_ = probChangeCube_[rateNo_][inStateNo_][inStateNo_] *
							             smaLiNd_->liNdCube_[local_addr + inStateNo_];
						else
							for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++)
								exStateLi_ += probChangeCube_[rateNo_][inStateNo_][exStateNo_] *
								              smaLiNd_->liNdCube_[local_addr + exStateNo_];
					}

					rateExPtnLi_ += greLiNd_->liNdCube_[local_addr + inStateNo_] * exStateLi_;
					//stateFrqArr_[inStateNo_] *
				}
			}

			double li_rate;

			if (!myrate.use_invar_site || rateNo_ < nRate_ - 1) {
				li_rate = rateExPtnLi_ * probRate_;
				(*ptn_rate)[ptnNo_] += li_rate * myrate.getRate(rateNo_);
			} else {
				li_rate = rateExPtnLi_ * myrate.prob_invar_site;
			}
			exPtnLi_ += li_rate;
		}

		if (exPtnLi_ <= 0.0) {
			exPtnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		(*ptn_rate)[ptnNo_] /= exPtnLi_;


	}

	/*
	int nSite_ = alignment.getNSite ();
	//  double totalRate_ = 0.0;
	for (int siteNo_ = 0; siteNo_ < nSite_; siteNo_ ++) {
		ptnNo_ = alignment.getPtn (siteNo_);
		double siteRate_ = (*ptn_rate)[ptnNo_];
		out << siteNo_ + 1 << "\t";
		out.width (10);
		out.precision (4);
		out.setf (ios::left);
		//    totalRate_ += siteRate_;

		out << siteRate_ << endl;
	}
	*/
}


//-----------------------------------------------
//-  codon-based Nielsen Yang Model
//-----------------------------------------------
extern double min_negative;

//#define DEBUG_LOGLI_CLASS

//#define DEBUG__LOGLI

/**
	compute the log likelihood of this tree via this an external branch in case of one rate
*/
double LiBr::cmpExLogLiSelection () {

	Seq *seq_;
	seq_ = &alignment.getSeq (smaNdNo_);

	int nState_ = mymodel.getNState ();
	int nClass_ = myrate.getNRate ();
	int class_state = nClass_ * nState_;

	int classNo_;
	double sum = 0;
	for (classNo_ = 0; classNo_ < nClass_; classNo_ ++) {
		if (classNo_ < nClass_ - 1)
			sum += mymodel.getClassProb(classNo_);
		cmpProbChangeFrq (len_, classNo_, probChangeCube_[classNo_]);
	}
	if (fabs(sum + mymodel.getClassProb(nClass_ - 1) - 1.0) > ZERO) {
		cout << "sum = " << sum + mymodel.getClassProb(nClass_ - 1) << "!!!!" << endl;
		((CodonNY98*) mymodel.model_)->nsSyProbVec[nClass_-1] = 1.0 - sum;
	}

	double logLi_ = 0.0;
	int nPtn_ = ptnlist.getNPtn ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	#pragma omp parallel for  private(classNo_)  \
	reduction(+: logLi_)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
		int siteNo_ = ptnlist.siteArr_ [ptnNo_];
		int exStateNo_ = (*seq_).items_[siteNo_];
		LDOUBLE exPtnLi_ = 0.0;
		LDOUBLE rateExPtnLi_ = 0.0;
		int start_addr = ptnNo_ * class_state;

		for (classNo_ = 0; classNo_ < nClass_; classNo_ ++) {
			int local_addr = start_addr + classNo_ * nState_;

			rateExPtnLi_ = 0.0;
			if (exStateNo_ < nState_) {
				for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
					rateExPtnLi_ += greLiNd_->liNdCube_[local_addr + inStateNo_] *
					                probChangeCube_[classNo_][exStateNo_][inStateNo_];

				}

				//rateExPtnLi_ *= stateFrqArr_[exStateNo_];
			} else {
				for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
					LDOUBLE exStateLi_ = 0.0;

					if (exStateNo_ == BS_UNKNOWN)
						exStateLi_ = mymodel.stateFrqArr_[inStateNo_];
					else {
						for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++)
							exStateLi_ += probChangeCube_[classNo_][inStateNo_][exStateNo_] *
							              smaLiNd_->liNdCube_[local_addr + exStateNo_];
					}

					rateExPtnLi_ += greLiNd_->liNdCube_[local_addr + inStateNo_] * exStateLi_;
					//stateFrqArr_[inStateNo_] *
				}
			}


			exPtnLi_ += rateExPtnLi_ * mymodel.getClassProb(classNo_);

		}

		if (exPtnLi_ <= 0.0) {
			exPtnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double logExPtnLi_ = cmpLogLi (exPtnLi_);
		logLi_ += logExPtnLi_ * ptnlist.weightArr_[ptnNo_];
		//std::cout << ptnNo_ << " / " << logLi_ << endl;
	}

	logLi_ += greLiNd_->liscale + smaLiNd_->liscale;
	return logLi_;

}

/**
	compute the log likelihood of this tree via this an external branch in case of one rate
*/
double LiBr::cmpExLogLiDerivativeSelection (double &logli_derv1_ret, double
&logli_derv2_ret, bool calc_logli) {

	Seq *seq_;
	seq_ = &alignment.getSeq (smaNdNo_);

	int nState_ = mymodel.getNState ();
	int nClass_ = myrate.getNRate ();
	int class_state = nClass_ * nState_;

	int classNo_;
	double sum = 0;
	for (classNo_ = 0; classNo_ < nClass_; classNo_ ++) {
		if (classNo_ < nClass_ - 1)
			sum += mymodel.getClassProb(classNo_);
		cmpProbChangeDerivativesFrq (len_, classNo_, probChangeCube_[classNo_],
		                             prob_cube_derv1[classNo_], prob_cube_derv2[classNo_]);
	}
	if (fabs(sum + mymodel.getClassProb(nClass_ - 1) - 1.0) > ZERO) {
		cout << "sum = " << sum + mymodel.getClassProb(nClass_ - 1) << "!!!!" << endl;
		((CodonNY98*) mymodel.model_)->nsSyProbVec[nClass_-1] = 1.0 - sum;
	}

	double logli = 0.0;
	double logli_derv1 = 0.0;
	double logli_derv2 = 0.0;
	int nPtn_ = ptnlist.getNPtn ();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);

	#pragma omp parallel for  private(classNo_)  \
	reduction(+: logli, logli_derv1, logli_derv2)
	for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
		int siteNo_ = ptnlist.siteArr_ [ptnNo_];
		int exStateNo_ = (*seq_).items_[siteNo_];
		LDOUBLE exPtnLi_ = 0.0;
		LDOUBLE exPtnLi_1 = 0.0;
		LDOUBLE exPtnLi_2 = 0.0;
		LDOUBLE rateExPtnLi_ = 0.0;
		LDOUBLE rateExPtnLi_1 = 0.0;
		LDOUBLE rateExPtnLi_2 = 0.0;
		int start_addr = ptnNo_ * class_state;

		for (classNo_ = 0; classNo_ < nClass_; classNo_ ++) {
			int local_addr = start_addr + classNo_ * nState_;

			rateExPtnLi_ = 0.0;
			rateExPtnLi_1 = 0.0;
			rateExPtnLi_2 = 0.0;
			if (exStateNo_ < nState_) {
				for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
					rateExPtnLi_ += greLiNd_->liNdCube_[local_addr + inStateNo_] *
					                probChangeCube_[classNo_][exStateNo_][inStateNo_];
					rateExPtnLi_1 += greLiNd_->liNdCube_[local_addr + inStateNo_] *
					                 prob_cube_derv1[classNo_][exStateNo_][inStateNo_];
					rateExPtnLi_2 += greLiNd_->liNdCube_[local_addr + inStateNo_] *
					                 prob_cube_derv2[classNo_][exStateNo_][inStateNo_];

				}

				//rateExPtnLi_ *= stateFrqArr_[exStateNo_];
			} else {
				for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
					LDOUBLE exStateLi_ = 0.0;
					LDOUBLE exStateLi_1 = 0.0;
					LDOUBLE exStateLi_2 = 0.0;

					if (exStateNo_ == BS_UNKNOWN)
						exStateLi_ = mymodel.stateFrqArr_[inStateNo_];
					else {
						for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++) {
							exStateLi_ += probChangeCube_[classNo_][inStateNo_][exStateNo_] *
							              smaLiNd_->liNdCube_[local_addr + exStateNo_];
							exStateLi_1 += prob_cube_derv1[classNo_][inStateNo_][exStateNo_] *
							               smaLiNd_->liNdCube_[local_addr + exStateNo_];
							exStateLi_2 += prob_cube_derv2[classNo_][inStateNo_][exStateNo_] *
							               smaLiNd_->liNdCube_[local_addr + exStateNo_];
						}
					}

					rateExPtnLi_ += greLiNd_->liNdCube_[local_addr + inStateNo_] * exStateLi_;
					rateExPtnLi_1 += greLiNd_->liNdCube_[local_addr + inStateNo_] * exStateLi_1;
					rateExPtnLi_2 += greLiNd_->liNdCube_[local_addr + inStateNo_] * exStateLi_2;
					//stateFrqArr_[inStateNo_] *
				}
			}


			exPtnLi_ += rateExPtnLi_ * mymodel.getClassProb(classNo_);
			exPtnLi_1 += rateExPtnLi_1 * mymodel.getClassProb(classNo_);
			exPtnLi_2 += rateExPtnLi_2 * mymodel.getClassProb(classNo_);

		}

		if (exPtnLi_ <= 0.0) {
			exPtnLi_ = VERY_SMALL_POSITIVE;
			Utl::announceError (BAD_CMP_LI);
		}

		double temp1 = exPtnLi_1 / exPtnLi_;
		double temp2 = exPtnLi_2 / exPtnLi_;
		logli_derv1 += temp1 * ptnlist.weightArr_[ptnNo_];
		logli_derv2 += (temp2 - temp1 * temp1) * ptnlist.weightArr_[ptnNo_];
		if (calc_logli)
			logli += cmpLogLi(exPtnLi_) * ptnlist.weightArr_[ptnNo_];
		//std::cout << ptnNo_ << " / " << logLi_ << endl;
	}
	logli_derv1_ret = logli_derv1;
	logli_derv2_ret = logli_derv2;
	logli += greLiNd_->liscale + smaLiNd_->liscale;
	return logli;
}



/**
	compute the log likelihood of this tree via this an external 	branch in case of one rate
*/

void LiBr::empiricalBayesAnalysis(ostream &out) {

	if (!isMasterProc())
		return;

	cout << "Inferring sites under positive selection..." << endl;

	int classNo_;
	int nState_ = mymodel.getNState ();
	int nClass_ = myrate.getNRate ();
	int class_state = nClass_ * nState_;
	int siteNo_;
	char line[100];

	out << endl << "Positively selected codon sites (*: P>95%; **: P>99%)" << endl;
	out << " Site   Pr(w>1)     Mean w" << endl;
	
	double threshold0 = 0.5;
	double threshold1 = 0.95;
	double threshold2 = 0.99;


	Seq *seq_;
	seq_ = &alignment.getSeq (smaNdNo_);


	double sum = 0;
	for (classNo_ = 0; classNo_ < nClass_; classNo_ ++) {
		if (classNo_ < nClass_ - 1)
			sum += mymodel.getClassProb(classNo_);
		cmpProbChangeFrq (len_, classNo_, probChangeCube_[classNo_]);
	}
	if (fabs(sum + mymodel.getClassProb(nClass_ - 1) - 1.0) > ZERO) {
		cout << "sum = " << sum + mymodel.getClassProb(nClass_ - 1) << "!!!!" << endl;
		((CodonNY98*) mymodel.model_)->nsSyProbVec[nClass_-1] = 1.0 - sum;
	}

	//double logLi_ = 0.0;
	//int nPtn_ = ptnlist.getNPtn ();
	int nSite_ = alignment.getNSite();

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);


	LDOUBLE *rateExPtnLi_ = new LDOUBLE [nClass_];
	double *logExPtnLi_ = new double [nSite_];
	double *mean_w = new double [nSite_];
	double *pos_probability = new double[nSite_];

	for (siteNo_ = 0; siteNo_ < nSite_; siteNo_ ++) {

		//for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
		//int siteNo_ = ptnlist.siteArr_ [ptnNo_];
		int ptnNo_ = alignment.getPtn(siteNo_);
		int exStateNo_ = (*seq_).items_[siteNo_];
		LDOUBLE exPtnLi_ = 0.0;
		LDOUBLE max_value = 0.0;
		int start_addr = ptnNo_ * class_state;

		for (classNo_ = 0; classNo_ < nClass_; classNo_ ++) {
			int local_addr = start_addr + classNo_ * nState_;
			rateExPtnLi_[classNo_] = 0.0;
			if (exStateNo_ < nState_) {
				for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
					rateExPtnLi_[classNo_] += greLiNd_->liNdCube_[local_addr + inStateNo_] *
					                          probChangeCube_[classNo_][exStateNo_][inStateNo_];

				}

				//rateExPtnLi_ *= stateFrqArr_[exStateNo_];
			} else {
				for (int inStateNo_ = 0; inStateNo_ < nState_; inStateNo_ ++) {
					LDOUBLE exStateLi_ = 0.0;

					if (exStateNo_ == BS_UNKNOWN)
						exStateLi_ = mymodel.stateFrqArr_[inStateNo_];
					else {
						for (int exStateNo_ = 0; exStateNo_ < nState_; exStateNo_ ++)
							exStateLi_ += probChangeCube_[classNo_][inStateNo_][exStateNo_] *
							              smaLiNd_->liNdCube_[local_addr + exStateNo_];
					}

					rateExPtnLi_[classNo_] += greLiNd_->liNdCube_[local_addr + inStateNo_] * exStateLi_;
					//stateFrqArr_[inStateNo_] *
				}
			}

			rateExPtnLi_[classNo_] *= mymodel.getClassProb(classNo_);
			if (mymodel.getClassRatio(classNo_) > 1.0) {
				max_value += rateExPtnLi_[classNo_];
			}
			exPtnLi_ += rateExPtnLi_[classNo_];

		}

		if (exPtnLi_ < -ZERO) {
			cout << "len = " << len_ << endl;
			Utl::announceError (BAD_CMP_LI);
		}

		logExPtnLi_[siteNo_] = cmpLogLi (exPtnLi_);


		pos_probability[siteNo_] = max_value / exPtnLi_;
		mean_w[siteNo_] = 0.0;
		for (int cNo = 0; cNo < nClass_; cNo++)
			mean_w[siteNo_] += (rateExPtnLi_[cNo] / exPtnLi_) * mymodel.getClassRatio(cNo);
			
		if (pos_probability[siteNo_] > threshold0) {
			sprintf(line, "%5d %8.3f", alignment.getOriginSite(siteNo_) + 1, pos_probability[siteNo_]);
			out << line;
			if (pos_probability[siteNo_] > threshold1)
				out << "*";
			else
				out << " ";
			if (pos_probability[siteNo_] > threshold2)
				out << "*";
			else
				out << " ";
			sprintf(line, "%10.3f", mean_w[siteNo_]);

			out << line << endl;

		}
		//std::cout << ptnNo_ << " / " << logLi_ << endl;
	}


	out << endl << "Detailed information for all codon sites";
	if (alignment.getOriginSite(nSite_-1) != nSite_-1) {
		out << " (Some sites were discarded due to gaps/stop codons)";
	}
	out << ":" << endl;
	out << " Site   Pr(w>1)     Mean w     LogL" << endl;


	for (siteNo_ = 0; siteNo_ < nSite_; siteNo_ ++) {
		sprintf(line, "%5d %8.3f %11.3f %16.10f", alignment.getOriginSite(siteNo_) + 1, 
			pos_probability[siteNo_], mean_w[siteNo_], logExPtnLi_[siteNo_]);
		out << line << endl;
		
	}

	delete mean_w;
	delete logExPtnLi_;
	delete rateExPtnLi_;
	
}


/**********************************************************
*
*    General functions to compute log likelihood and derivatives
*
**********************************************************/

	/**
		compute the log-likelihood from external branch
		@return tree log-likelihood
	*/

double LiBr::cmpExLogLi () {

	RATE_TYPE rateType_ = myrate.getType ();

#ifdef PARALLEL
	/*
	if (opt_urtree.isParallelPhase(PARALLEL_LOGLI)) {
		if (myrate.isNsSyHeterogenous())
			return cmpExLogLiSelection();
		else
			if (rateType_ == UNIFORM || (rateType_ == SITE_SPECIFIC && myrate.isOptedSpecificRate () == 0))
				return par_cmpExLogLiUniformRate ();
			else
				if (rateType_ == SITE_SPECIFIC)
					return cmpExLogLiSpecificRate ();
				else
					return cmpExLogLiGammaRate ();

	}
	*/
#endif // PARALLEL

	if (myrate.isNsSyHeterogenous())
		return cmpExLogLiSelection();
	else
		if (rateType_ == UNIFORM || (rateType_ == SITE_SPECIFIC && myrate.isOptedSpecificRate () == 0))
			return cmpExLogLiUniformRate ();
		else
			if (rateType_ == SITE_SPECIFIC)
				return cmpExLogLiSpecificRate ();
			else
				return cmpExLogLiGammaRate ();
}


// compute the log-likelihood from external branchd
double LiBr::cmpExLogLiDerivatives (double &logli_derv1, double &logli_derv2, bool calc_logli) {

	RATE_TYPE rateType_ = myrate.getType ();
	if (myrate.isNsSyHeterogenous())
		return cmpExLogLiDerivativeSelection(logli_derv1, logli_derv2, calc_logli);
	else
		if (rateType_ == UNIFORM || (rateType_ == SITE_SPECIFIC && myrate.isOptedSpecificRate () == 0))
			return cmpExLogLiDerivativeUniformRate(logli_derv1, logli_derv2, calc_logli);

		else
			if (rateType_ == SITE_SPECIFIC)
				return cmpExLogLiDerivativeSpecificRate (logli_derv1, logli_derv2, calc_logli);
			else
				return cmpExLogLiDerivativeGammaRate (logli_derv1, logli_derv2, calc_logli);

}

// compute the log-likelihood from internal branch
double LiBr::cmpInLogLi () {
	RATE_TYPE rateType_ = myrate.getType ();

#ifdef PARALLEL
	/*
	if (opt_urtree.isParallelPhase(PARALLEL_LOGLI)) {
		if (myrate.isNsSyHeterogenous())
			return cmpInLogLiSelection();
		else
			if (rateType_ == UNIFORM || (rateType_ == SITE_SPECIFIC && myrate.isOptedSpecificRate () == 0) )
				return par_cmpInLogLiUniformRate ();
			else
				if (rateType_ == SITE_SPECIFIC)
					// not done
					return cmpInLogLiSpecificRate ();
				else
					// not done
					return cmpInLogLiGammaRate ();
	}
	*/
#endif // PARALLEL

	if (myrate.isNsSyHeterogenous())
		return cmpInLogLiSelection();
	else
		if (rateType_ == UNIFORM || (rateType_ == SITE_SPECIFIC && myrate.isOptedSpecificRate () == 0) )
			return cmpInLogLiUniformRate ();
		else
			if (rateType_ == SITE_SPECIFIC)
				return cmpInLogLiSpecificRate ();
			else
				return cmpInLogLiGammaRate ();
}


// compute the log-likelihood derivatives from internal branch
double LiBr::cmpInLogLiDerivatives (double &logli_derv1, double &logli_derv2, bool calc_logli) {
	RATE_TYPE rateType_ = myrate.getType ();
	if (myrate.isNsSyHeterogenous())
		return cmpInLogLiDerivativeSelection(logli_derv1, logli_derv2, calc_logli);
	else
		if (rateType_ == UNIFORM || (rateType_ == SITE_SPECIFIC && myrate.isOptedSpecificRate () == 0) )
			return cmpInLogLiDerivativeUniformRate (logli_derv1, logli_derv2, calc_logli);
		else
			if (rateType_ == SITE_SPECIFIC)
				return cmpInLogLiDerivativeSpecificRate (logli_derv1, logli_derv2, calc_logli);
			else
				return cmpInLogLiDerivativeGammaRate (logli_derv1, logli_derv2, calc_logli);
}


//-------------------------------------------------------------------
//set the two liNds to be not cmped
void LiBr::openLiNd () {
	greLiNd_->turnOffIsCmped ();
	smaLiNd_->turnOffIsCmped ();
}

//-------------------------------------------------------------------
//set the two liNds to be not cmped
void LiBr::openTailLiNd () {
	if (getTailNd () == greNdNo_ )
		greLiNd_->turnOffIsCmped ();
	else
		smaLiNd_->turnOffIsCmped ();
}

//-------------------------------------------------------------------
//set the two liNds to be not cmped
void LiBr::openHeadLiNd () {
	if (getHeadNd () == greNdNo_ )
		greLiNd_->turnOffIsCmped ();
	else
		smaLiNd_->turnOffIsCmped ();
}


//======================================================
//set memory for this liBr

void LiBr::setLimit () {}

//======================================================
//release all memory of this class
void LiBr::release () {
	smaLiNd_->release ();
	greLiNd_->release ();
}

//======================================================
//the destructor
LiBr::~LiBr () {
	release ();
}
