/***************************************************************************
 *                                                                         *
 *                  (begin: Feb 20 2003)                                   *
 *                                                                         *
 *   Parallel IQPNNI - Important Quartet Puzzle with NNI                   *
 *                                                                         *
 *   Copyright (C) 2005 by Le Sy Vinh, Bui Quang Minh, Arndt von Haeseler  *
 *   Copyright (C) 2003-2004 by Le Sy Vinh, Arndt von Haeseler             *
 *   {vinh,minh}@cs.uni-duesseldorf.de                                     *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

// innd.cpp: implementation of the inNd class.
//
//////////////////////////////////////////////////////////////////////

#include <iostream>
#include <string.h>

#include "outstream.h"
#include "lind.h"
#include "opturtree.h"
#include "model.h"
#include "ptnls.h"
#include "rate.h"
#include "ptnratecube.h"

const double SCALING_THRESHOLD = 1e-150;

//#define TEST_NUMBER_DIFF

#ifdef TEST_NUMBER_DIFF
double min_num_diff = 1.0;
#endif

//#define PRINT_NEGATIVE_INFO

//--------------------------------------------------------------------
//the constructor function
LiNd::LiNd () {
	isCmped_ = 0;
	isEx_ = -1;

	int nPtn_ = ptnlist.getNPtn ();

	int nRate_ = myrate.getNRate ();
	int nState_ = mymodel.getNState ();
	Utl::getMem (liNdCube_, nRate_ * nState_ * nPtn_);
//	Utl::getMem(liscale, nRate_, nPtn_);
}


//--------------------------------------------------------------------
//set the id, ndType, parent node, and branch no for this node
void LiNd::set (int id, EX_IN ndType, int parNdNo, int parBrNo) {
	id_ = id;
	if (ndType == EX)
		isEx_ = 1;
	else
		isEx_ = 0;

	parNdNo_ = parNdNo;
	parBrNo_ = parBrNo;

	isExDesNdCreated_ = 0;
	isCmped_ = 0;
}


//--------------------------------------------------------------------
void LiNd::turnOffIsCmped () {
	isCmped_ = 0;
}

//--------------------------------------------------------------------
int LiNd::getId () {
	return id_;
}

Vec<int> &LiNd::getExDesNd () {
	return exDesNdNoLs_;
}

//--------------------------------------------------------------------
//the the parent node, and branch no for this node
void LiNd::getPar (int &parNdNo, int &parBrNo) {
	parNdNo = parNdNo_;
	parBrNo = parBrNo_;
}

void LiNd::setParNdBr (int newParNdNo, int newParBrNo) {
	parNdNo_ = newParNdNo;
	parBrNo_ = newParBrNo;
}

void LiNd::writeInf () {
	std::cout << this << " / " << id_ << " / " << parNdNo_ << " / " << parBrNo_ << " / " << endl;
}

//-------------------------------------------------------------------
//compute the likehood of (ptnNo, rateNo, stateNo) via the parBrNo_
//LDOUBLE LiNd::cmpLi (int ptnNo, int rateNo, int stateNo) {
//	if (isCmped_ == 0)
//		cmpLi ();
//	return liNdCube_[rateNo][stateNo][ptnNo];
//} //end of function


//-------------------------------------------------------------------
//reCompute the likelihood for all ptnNo, rateNo, stateNo at this node
void LiNd::reCmpLi () {
	isCmped_ = 0;
	cmpLi ();
}

DMat20 proChangeMat1_;
DMat20 proChangeMat2_;


//-------------------------------------------------------------------
//compute the likelihood for all ptnNo, rateNo, stateNo at this node
void LiNd::cmpLiUniformRate () {
	if (isCmped_ == 1)
		return;

	//we do not need to reCmp again
	//DVec20 stateFrqArr_;


	//mymodel.getStateFrq (stateFrqArr_);
	//then, compute the likelihood of this internal node
	int nState_ = mymodel.getNState ();
	//    std::cout << id_ << endl;


	if (isEx_ == 1) { //the case of external case
		Seq *seq_;
		seq_ = &alignment.getSeq (id_);
		int nPtn_ = ptnlist.getNPtn ();

		liscale = 0.0;

		#pragma omp parallel for  
		for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {

			// get the starting address of the li-cube
			PVec lipart = liNdCube_ + ptnNo_ * nState_;

			int siteNo_ = ptnlist.siteArr_[ptnNo_];
			int ndStateNo_ = (*seq_).items_[siteNo_];

			if (ndStateNo_ < nState_) {
				
				//for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
				//	liNdCube_[0][stateNo_][ptnNo_] = 0.0;
				
				memset(lipart, 0, sizeof(LDOUBLE) * nState_);
				lipart[ndStateNo_] = 1.0;

			} else {
				for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
					lipart[stateNo_] = 1.0;

				if (ndStateNo_ != BS_UNKNOWN && mymodel.getDataType () == NUCLEOTIDE)
					switch (ndStateNo_) {
					case BASE_AC:
 						lipart[BASE_G] = 0.0;
						lipart[BASE_T] = 0.0;
						break;
					case BASE_AG:
						lipart[BASE_C] = 0.0;
						lipart[BASE_T] = 0.0;
						break;
					case BASE_AT:
						lipart[BASE_C] = 0.0;
						lipart[BASE_G] = 0.0;
						break;
					case BASE_CG:
						lipart[BASE_A] = 0.0;
						lipart[BASE_T] = 0.0;
						break;
					case BASE_CT:
						lipart[BASE_A] = 0.0;
						lipart[BASE_G] = 0.0;
						break;
					case BASE_GT:
						lipart[BASE_A] = 0.0;
						lipart[BASE_C] = 0.0;
						break;
					case BASE_CGT:
						lipart[BASE_A] = 0.0;
						break;
					case BASE_AGT:
						lipart[BASE_C] = 0.0;
						break;
					case BASE_ACT:
						lipart[BASE_G] = 0.0;
						break;
					case BASE_ACG:
						lipart[BASE_T] = 0.0;
						break;
					} // end of switch
			} //end of else
		} //end of for

	} else { //the case of internal node

		//first compute the likelihood for all childrent
		int chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_;

		opt_urtree.inNdArr_[id_].get2RemNeiNdBr (parNdNo_, chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_);
		double chiLen1_ = opt_urtree.getBrLen (chiBrNo1_);
		double chiLen2_ = opt_urtree.getBrLen (chiBrNo2_);

		int nState_ = mymodel.getNState ();

		LiNd *chiLi1_;

		chiLi1_ = &opt_urtree.liBrArr_[chiBrNo1_].getLiNd (chiNdNo1_);

		chiLi1_->cmpLiUniformRate ();

		LiNd *chiLi2_;
		chiLi2_ = &opt_urtree.liBrArr_[chiBrNo2_].getLiNd (chiNdNo2_);
		chiLi2_->cmpLiUniformRate ();

		mymodel.cmpProbChange (chiLen1_, proChangeMat1_);
		mymodel.cmpProbChange (chiLen2_, proChangeMat2_);
		
		int nPtn_ = ptnlist.getNPtn ();

		// initialize liscale = sum of childrens liscale
		liscale = chiLi1_->liscale + chiLi2_->liscale;

		#pragma omp parallel for  
		for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {

			int stateNo_;
			bool should_scale = false;
			int start_addr = ptnNo_ * nState_;
			// get the starting address of the li-cube
			PVec lipart = liNdCube_ + start_addr;
			PVec chi1_lipart = chiLi1_->liNdCube_ + start_addr;
			PVec chi2_lipart = chiLi2_->liNdCube_ + start_addr;


			for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) {

				int chiStateNo_;
				LDOUBLE chiLiBr1_ = 0.0;
				LDOUBLE chiLiBr2_ = 0.0;
				for (chiStateNo_ = 0; chiStateNo_ < nState_; chiStateNo_ ++) {

					chiLiBr1_ += chi1_lipart[chiStateNo_] *
					             proChangeMat1_[stateNo_][chiStateNo_];
					chiLiBr2_ += chi2_lipart[chiStateNo_] *
					             proChangeMat2_[stateNo_][chiStateNo_];
				}

				lipart[stateNo_] = chiLiBr1_ * chiLiBr2_;

				if (lipart[stateNo_] < SCALING_THRESHOLD && lipart[stateNo_] > 0.0)
					should_scale = true;

			}//end of for stateNo_

#ifdef TEST_NUMBER_DIFF
			LDOUBLE max_state_li = lipart[0];
			LDOUBLE min_state_li = 1.0;
			for (stateNo_ = 1; stateNo_ < nState_; stateNo_ ++) 
				if (max_state_li < lipart[stateNo_]) 
					max_state_li = lipart[stateNo_];
			for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) 
				if (lipart[stateNo_] > 0.0 && min_state_li > lipart[stateNo_]) 
					min_state_li = lipart[stateNo_];
			double diff = min_state_li / max_state_li;
			if (min_num_diff > diff) {
				min_num_diff = diff;
				cout << "likelihood different = " << min_num_diff << endl;
			}
#endif


			if (!should_scale) continue;
			/* now do the scaling */

			LDOUBLE limax = lipart[0];
			for (stateNo_ = 1; stateNo_ < nState_; stateNo_ ++) 
				if (limax < lipart[stateNo_]) 
					limax = lipart[stateNo_];
			
			for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) 
				lipart[stateNo_] /= limax;

			// sum over pattern counts
			liscale += log(limax) * ptnlist.weightArr_[ptnNo_];

		} //end of for ptnNo_

	} // end of else

	isCmped_ = 1;
}

//---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//---------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------
//compute the likelihood for all ptnNo, rateNo, stateNo at this node
void LiNd::cmpLiSpecificRate () {
	if (isCmped_ == 1)
		return;

	//we do not need to reCmp again
	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);
	//then, compute the likelihood of this internal node
	int nState_ = mymodel.getNState ();
	//    std::cout << id_ << endl;


	if (isEx_ == 1) { //the case of external case
		Seq *seq_;
		seq_ = &alignment.getSeq (id_);
		int nPtn_ = ptnlist.getNPtn ();

		liscale = 0.0;

		#pragma omp parallel for  
		for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
			int siteNo_ = ptnlist.siteArr_[ptnNo_];
			int ndStateNo_ = (*seq_).items_[siteNo_];
			PVec lipart = liNdCube_ + ptnNo_ * nState_;

			if (ndStateNo_ < nState_) {
				for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
					lipart[stateNo_] = 0.0;

				lipart[ndStateNo_] = 1.0;

			} else {
				for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
					lipart[stateNo_] = 1.0;

				if (ndStateNo_ != BS_UNKNOWN && mymodel.getDataType () == NUCLEOTIDE)
					switch (ndStateNo_) {
					case BASE_AC:
						lipart[BASE_G] = 0.0;
						lipart[BASE_T] = 0.0;
						break;
					case BASE_AG:
						lipart[BASE_C] = 0.0;
						lipart[BASE_T] = 0.0;
						break;
					case BASE_AT:
						lipart[BASE_C] = 0.0;
						lipart[BASE_G] = 0.0;
						break;
					case BASE_CG:
						lipart[BASE_A] = 0.0;
						lipart[BASE_T] = 0.0;
						break;
					case BASE_CT:
						lipart[BASE_A] = 0.0;
						lipart[BASE_G] = 0.0;
						break;
					case BASE_GT:
						lipart[BASE_A] = 0.0;
						lipart[BASE_C] = 0.0;
						break;
					case BASE_CGT:
						lipart[BASE_A] = 0.0;
						break;

					case BASE_AGT:
						lipart[BASE_C] = 0.0;
						break;
					case BASE_ACT:
						lipart[BASE_G] = 0.0;
						break;
					case BASE_ACG:
						lipart[BASE_T] = 0.0;
						break;
					} // end of switch
			} //end of else
		} //end of for

	} else { //the case of internal node

		//first compute the likelihood for all childrent
		int chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_;

		opt_urtree.inNdArr_[id_].get2RemNeiNdBr (parNdNo_, chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_);

		LiNd *chiLi1_;

		chiLi1_ = &opt_urtree.liBrArr_[chiBrNo1_].getLiNd (chiNdNo1_);

		chiLi1_->cmpLiSpecificRate ();

		LiNd *chiLi2_;
		chiLi2_ = &opt_urtree.liBrArr_[chiBrNo2_].getLiNd (chiNdNo2_);
		chiLi2_->cmpLiSpecificRate ();

		int nState_ = mymodel.getNState ();
		int nPtn_ = ptnlist.getNPtn ();

		double chiLen1_ = opt_urtree.getBrLen (chiBrNo1_);
		double chiLen2_ = opt_urtree.getBrLen (chiBrNo2_);

		Vec<double> *ptnRateArr_;
		ptnRateArr_ = myrate.getPtnRate ();

		// initialize liscale = sum of childrens liscale
		liscale = chiLi1_->liscale + chiLi2_->liscale;

		#pragma omp parallel for   \
		private(proChangeMat1_,	proChangeMat2_)
		for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
			double ptnRate_ = (*ptnRateArr_)[ptnNo_];
			int stateNo_;
			bool should_scale = false;
			int start_addr = ptnNo_ * nState_;

			PVec lipart = liNdCube_ + start_addr;			
			PVec chi1_lipart = chiLi1_->liNdCube_ + start_addr;			
			PVec chi2_lipart = chiLi2_->liNdCube_ + start_addr;			

			mymodel.cmpProbChange (chiLen1_ * ptnRate_, proChangeMat1_);
			mymodel.cmpProbChange (chiLen2_ * ptnRate_, proChangeMat2_);

			for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) {

				int chiStateNo_;
				LDOUBLE chiLiBr1_ = 0.0;
				LDOUBLE chiLiBr2_ = 0.0;
				
				for (chiStateNo_ = 0; chiStateNo_ < nState_; chiStateNo_ ++) {
					chiLiBr1_ += chi1_lipart[chiStateNo_] *
					             proChangeMat1_[stateNo_][chiStateNo_];

					chiLiBr2_ += chi2_lipart[chiStateNo_] *
					             proChangeMat2_[stateNo_][chiStateNo_];
				}

				lipart[stateNo_] = chiLiBr1_ * chiLiBr2_;
				if (lipart[stateNo_] < SCALING_THRESHOLD && lipart[stateNo_] > 0.0)
					should_scale = true;

			}//end of for stateNo_

			if (!should_scale) continue;
			/* now do the scaling */

			LDOUBLE limax = lipart[0];
			for (stateNo_ = 1; stateNo_ < nState_; stateNo_ ++) 
				if (limax < lipart[stateNo_]) 
					limax = lipart[stateNo_];
			
			for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) 
				lipart[stateNo_] /= limax;

			// sum over pattern counts
			liscale += log(limax) * ptnlist.weightArr_[ptnNo_];

		} //end of for ptnNo_

	} // end of else

	isCmped_ = 1;
}


DMat20 proChangeCube1_[MAX_NUM_RATE];
DMat20 proChangeCube2_[MAX_NUM_RATE];

//-------------------------------------------------------------------
//compute the likelihood for all ptnNo, rateNo, stateNo at this node
void LiNd::cmpLiGammaRate () {
	//we do not need to reCmp again
	if (isCmped_ == 1)
		return;


	opt_urtree.nCmpLi_ ++;

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);
	//then, compute the likelihood of this internal node
	int nRate_ = myrate.getNRate ();
	int nState_ = mymodel.getNState ();
	int rate_state = nRate_ * nState_;
	//double probRate_ = myrate.getProb ();

	//std::cout << id_ << endl;
	if (isEx_ == 1) { //the case of external case
		//    std::cout << id_ << endl;
		Seq *seq_;

		seq_ = &alignment.getSeq (id_);
		int nPtn_ = ptnlist.getNPtn ();
		
		liscale = 0.0;

		#pragma omp parallel for  
		for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
			int siteNo_ = ptnlist.siteArr_[ptnNo_];
			int ndStateNo_ = (*seq_).items_[siteNo_];

			int start_addr = ptnNo_ * rate_state;

			if (ndStateNo_ < nState_)
				for (int rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
					PVec lipart = liNdCube_ + (start_addr + rateNo_ * nState_);

					for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
						lipart[stateNo_] = 0.0;

					lipart[ndStateNo_] = 1.0;

				}
			else
				for (int rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
					PVec lipart = liNdCube_ + (start_addr + rateNo_ * nState_);

					for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
						lipart[stateNo_] = 1.0;

					if (ndStateNo_ != BS_UNKNOWN && mymodel.getDataType () == NUCLEOTIDE)
						switch (ndStateNo_) {
						case BASE_AC:
							lipart[BASE_G] = 0.0;
							lipart[BASE_T] = 0.0;
							break;
						case BASE_AG:
							lipart[BASE_C] = 0.0;
							lipart[BASE_T] = 0.0;
							break;
						case BASE_AT:
							lipart[BASE_C] = 0.0;
							lipart[BASE_G] = 0.0;
							break;
						case BASE_CG:
							lipart[BASE_A] = 0.0;
							lipart[BASE_T] = 0.0;
							break;
						case BASE_CT:
							lipart[BASE_A] = 0.0;
							lipart[BASE_G] = 0.0;
							break;
						case BASE_GT:
							lipart[BASE_A] = 0.0;
							lipart[BASE_C] = 0.0;
							break;

						case BASE_CGT:
							lipart[BASE_A] = 0.0;
							break;
						case BASE_AGT:
							lipart[BASE_C] = 0.0;
							break;
						case BASE_ACT:
							lipart[BASE_G] = 0.0;
							break;
						case BASE_ACG:
							lipart[BASE_T] = 0.0;
							break;
						} // end of switch
				} //end of for rateNo

		}

	} else { //the case of internal node

		//first compute the likelihood for all childrent
		int chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_;

		opt_urtree.inNdArr_[id_].get2RemNeiNdBr (parNdNo_, chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_);
		double chiLen1_ = opt_urtree.getBrLen (chiBrNo1_);
		double chiLen2_ = opt_urtree.getBrLen (chiBrNo2_);

		int nState_ = mymodel.getNState ();
		int nRate_ = myrate.getNRate ();
		//    double probRate_ = myrate.getProb ();


		LiNd *chiLi1_;
		chiLi1_ = &opt_urtree.liBrArr_[chiBrNo1_].getLiNd (chiNdNo1_);
		chiLi1_->cmpLiGammaRate ();

		LiNd *chiLi2_;
		chiLi2_ = &opt_urtree.liBrArr_[chiBrNo2_].getLiNd (chiNdNo2_);
		chiLi2_->cmpLiGammaRate ();

		/*
		DMat20 *proChangeCube1_;
		proChangeCube1_ = new DMat20[nRate_];

		DMat20 *proChangeCube2_;
		proChangeCube2_ = new DMat20[nRate_];
		*/

		int rateNo_;
		for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
			double rate_ = myrate.getRate (rateNo_);
			mymodel.cmpProbChange (chiLen1_ * rate_, proChangeCube1_[rateNo_]);
			mymodel.cmpProbChange (chiLen2_ * rate_, proChangeCube2_[rateNo_]);
		}
		
		int nPtn_ = ptnlist.getNPtn ();

		// initialize liscale = sum of childrens liscale
		liscale = chiLi1_->liscale + chiLi2_->liscale;

		#pragma omp parallel for   \
		private(rateNo_)
		for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {

			int start_addr = ptnNo_ * rate_state;
			bool should_scale = false;

			for (rateNo_ = 0; rateNo_ < nRate_; rateNo_ ++) {
				int stateNo_;
				int local_addr = start_addr + rateNo_ * nState_;
				PVec lipart = liNdCube_ + local_addr;
				PVec chi1_lipart = chiLi1_->liNdCube_ + local_addr;
				PVec chi2_lipart = chiLi2_->liNdCube_ + local_addr;

				for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) {

					int chiStateNo_;
					LDOUBLE chiLiBr1_ = 0.0;
					LDOUBLE chiLiBr2_ = 0.0;
					if (myrate.use_invar_site && rateNo_ == nRate_ - 1) {
						chiLiBr1_ += chi1_lipart[stateNo_] *
						             proChangeCube1_[rateNo_][stateNo_][stateNo_];
						chiLiBr2_ += chi2_lipart[stateNo_] * 
							proChangeCube2_[rateNo_][stateNo_][stateNo_];
					} else
					for (chiStateNo_ = 0; chiStateNo_ < nState_; chiStateNo_ ++) {
						chiLiBr1_ += chi1_lipart[chiStateNo_] *
						             proChangeCube1_[rateNo_][stateNo_][chiStateNo_];
						chiLiBr2_ += chi2_lipart[chiStateNo_] * 
							proChangeCube2_[rateNo_][stateNo_][chiStateNo_];
					}


					lipart[stateNo_] = chiLiBr1_ * chiLiBr2_;
					if (lipart[stateNo_] < SCALING_THRESHOLD && lipart[stateNo_] > 0.0) 
						should_scale = true;

				}//end of for stateNo_

			} // end of for rateNo_

			if (!should_scale) continue;
			/* now do the scaling */

			PVec lipart = liNdCube_ + start_addr;
			int loopit;
			LDOUBLE limax = lipart[0];
			for (loopit = 1; loopit < rate_state; loopit ++) 
				if (limax < lipart[loopit]) 
					limax = lipart[loopit];
			
			for (loopit = 0; loopit < rate_state; loopit ++) 
				lipart[loopit] /= limax;

			liscale += log(limax) * ptnlist.weightArr_[ptnNo_];

		} //end of for ptnNo_

		//delete [] proChangeCube1_;
		//delete [] proChangeCube2_;

	} // end of else

	isCmped_ = 1;
}

void print_li(int classNo_, PVec lipart, PVec chi1_lipart, PVec chi2_lipart) {
	int nState_ = mymodel.getNState ();
	for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) {

		//int chiStateNo_;
		LDOUBLE chiLiBr1_ = 0.0;
		LDOUBLE chiLiBr2_ = 0.0;
		cout << stateNo_ << endl;
		for (int chiStateNo_ = 0; chiStateNo_ < nState_; chiStateNo_ ++) {

			LDOUBLE chiLiBr1_part = chi1_lipart[chiStateNo_] *
				proChangeCube1_[classNo_][stateNo_][chiStateNo_];
			LDOUBLE chiLiBr2_part = chi2_lipart[chiStateNo_] * 
				proChangeCube2_[classNo_][stateNo_][chiStateNo_];
			if (chiLiBr1_part < 0.0 || chiLiBr2_part < 0.0) {
				cout << chi1_lipart[chiStateNo_] << " * " << 
					proChangeCube1_[classNo_][stateNo_][chiStateNo_] << " = " << 
					chiLiBr1_part << " $ " << 
					chi2_lipart[chiStateNo_] << " * " << 
					proChangeCube2_[classNo_][stateNo_][chiStateNo_] << " = " << 
					chiLiBr2_part << endl;
			}
			chiLiBr1_ += chiLiBr1_part;
			chiLiBr2_ += chiLiBr2_part;
		}

		//cout << chiLiBr1_ <<"  "<< chiLiBr2_ << "  " << chiLiBr1_ * chiLiBr2_ << endl;
		
	}
	exit(1);
}

//-------------------------------------------------------------------
//compute the likelihood for all ptnNo, rateNo, stateNo at this node
void LiNd::cmpLiSelection () {
	//we do not need to reCmp again
	if (isCmped_ == 1)
		return;


	opt_urtree.nCmpLi_ ++;

	//DVec20 stateFrqArr_;
	//mymodel.getStateFrq (stateFrqArr_);
	//then, compute the likelihood of this internal node
	int nClass_ = myrate.getNRate ();
	int nState_ = mymodel.getNState ();
	int class_state = nClass_ * nState_;

	//std::cout << id_ << endl;
	if (isEx_ == 1) { //the case of external case
		//    std::cout << id_ << endl;
		Seq *seq_;

		seq_ = &alignment.getSeq (id_);
		int nPtn_ = ptnlist.getNPtn ();

		liscale = 0.0;

		#pragma omp parallel for  
		for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {
			int siteNo_ = ptnlist.siteArr_[ptnNo_];
			int ndStateNo_ = (*seq_).items_[siteNo_];
			int start_addr = ptnNo_ * nClass_ * nState_;

			if (ndStateNo_ < nState_)
				for (int classNo_ = 0; classNo_ < nClass_; classNo_ ++) {
					PVec lipart = liNdCube_ + (start_addr + classNo_ * nState_);
					/*
					for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
						liNdCube_[ptnNo_][classNo_][stateNo_] = 0.0;
					*/
					memset(lipart, 0, sizeof(LDOUBLE) * nState_);
					lipart[ndStateNo_] = 1.0;

				}
			else
				for (int classNo_ = 0; classNo_ < nClass_; classNo_ ++) {
					PVec lipart = liNdCube_ + (start_addr + classNo_ * nState_);

					for (int stateNo_ = 0; stateNo_ < nState_; stateNo_ ++)
						lipart[stateNo_] = 1.0;
				} //end of for rateNo

		}

	} else { //the case of internal node

		//first compute the likelihood for all childrent
		int chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_;

		opt_urtree.inNdArr_[id_].get2RemNeiNdBr (parNdNo_, chiNdNo1_, chiBrNo1_, chiNdNo2_, chiBrNo2_);
		double chiLen1_ = opt_urtree.getBrLen (chiBrNo1_);
		double chiLen2_ = opt_urtree.getBrLen (chiBrNo2_);

		int nState_ = mymodel.getNState ();
		int nClass_ = myrate.getNRate ();
		//    double probRate_ = myrate.getProb ();


		LiNd *chiLi1_;
		chiLi1_ = &opt_urtree.liBrArr_[chiBrNo1_].getLiNd (chiNdNo1_);
		chiLi1_->cmpLiSelection ();

		LiNd *chiLi2_;
		chiLi2_ = &opt_urtree.liBrArr_[chiBrNo2_].getLiNd (chiNdNo2_);
		chiLi2_->cmpLiSelection ();


		int classNo_;
		for (classNo_ = 0; classNo_ < nClass_; classNo_ ++) {
			mymodel.cmpProbChange (chiLen1_, classNo_, proChangeCube1_[classNo_]);
			mymodel.cmpProbChange (chiLen2_, classNo_, proChangeCube2_[classNo_]);
		}
		
		int nPtn_ = ptnlist.getNPtn ();

		// initialize liscale = sum of childrens liscale
		liscale = chiLi1_->liscale + chiLi2_->liscale;

		#pragma omp parallel for   \
		private(classNo_)
		for (int ptnNo_ = 0; ptnNo_ < nPtn_; ptnNo_ ++) {

			int start_addr = ptnNo_ * class_state;
			bool should_scale = false;

			for (classNo_ = 0; classNo_ < nClass_; classNo_++) {
				int stateNo_;
				int local_addr = start_addr + classNo_ * nState_;
				PVec lipart = liNdCube_ + local_addr;
				PVec chi1_lipart = chiLi1_->liNdCube_ + local_addr;
				PVec chi2_lipart = chiLi2_->liNdCube_ + local_addr;

				for (stateNo_ = 0; stateNo_ < nState_; stateNo_ ++) {

					int chiStateNo_;
					LDOUBLE chiLiBr1_ = 0.0;
					LDOUBLE chiLiBr2_ = 0.0;
					for (chiStateNo_ = 0; chiStateNo_ < nState_; chiStateNo_ ++) {
						chiLiBr1_ += chi1_lipart[chiStateNo_] *
						             proChangeCube1_[classNo_][stateNo_][chiStateNo_];
						chiLiBr2_ += chi2_lipart[chiStateNo_] * 
							proChangeCube2_[classNo_][stateNo_][chiStateNo_];
					}

					lipart[stateNo_] = chiLiBr1_ * chiLiBr2_;
#ifdef PRINT_NEGATIVE_INFO
					if (lipart[stateNo_] < 0.0) 
						print_li(classNo_, lipart, chi1_lipart, chi2_lipart);
#endif

					if (lipart[stateNo_] < SCALING_THRESHOLD && lipart[stateNo_] > 0.0) 
						should_scale = true;

				}//end of for stateNo_
			} // end of for classNo_

			if (!should_scale) continue;
			/* now do the scaling */

			//cout << "!" ;

			PVec lipart = liNdCube_ + start_addr;
			int loopit;
			LDOUBLE limax = lipart[0];
			for (loopit = 1; loopit < class_state; loopit ++) 
				if (limax < lipart[loopit]) 
					limax = lipart[loopit];
			
			for (loopit = 0; loopit < class_state; loopit ++) 
				lipart[loopit] /= limax;

			liscale += log(limax) * ptnlist.weightArr_[ptnNo_];


		} //end of for ptnNo_

		//delete [] proChangeCube1_;
		//delete [] proChangeCube2_;

	} // end of else

	isCmped_ = 1;
}


//compute the likelihood for all ptnNo, rateNo, baseNo at this node
void LiNd::cmpLi () {
	RATE_TYPE rateType_ = myrate.getType ();
	if (myrate.isNsSyHeterogenous())
		cmpLiSelection();
	else
	if (rateType_ == UNIFORM || (rateType_ == SITE_SPECIFIC && myrate.isOptedSpecificRate () == 0))
		cmpLiUniformRate ();
	else
		if (rateType_ == SITE_SPECIFIC)
			cmpLiSpecificRate ();
		else
			cmpLiGammaRate ();
}

//--------------------------------------------------------------------
//create the external descendant nodes no list of this node
void LiNd::createExDesNd (Vec<int> &exDesNdNoLs) {
	if (isExDesNdCreated_ == 0) {
		if (isEx_ == 1) {
			exDesNdNoLs_.set (1, 0);
			exDesNdNoLs_ += id_;
		} else {
			int chiBrNo1_, chiBrNo2_;
			opt_urtree.inNdArr_[id_].get2RemBr (parBrNo_, chiBrNo1_, chiBrNo2_);

			exDesNdNoLs_.set (MAX_NUM_SEQ, 0);

			Vec<int> chiExDesNdNoLs1_;
			opt_urtree.liBrArr_[chiBrNo1_].getRemLiNd (id_).createExDesNd (chiExDesNdNoLs1_);
			exDesNdNoLs_ += chiExDesNdNoLs1_;

			Vec<int> chiExDesNdNoLs2_;
			opt_urtree.liBrArr_[chiBrNo2_].getRemLiNd (id_).createExDesNd (chiExDesNdNoLs2_);
			exDesNdNoLs_ += chiExDesNdNoLs2_;
		}
	}


	exDesNdNoLs = exDesNdNoLs_;
	isExDesNdCreated_ = 1;
}

//--------------------------------------------------------------------
//create the external descendant nodes no list of this node
void LiNd::createExDesNd () {
	createExDesNd (exDesNdNoLs_);
}

//--------------------------------------------------------------------
//set memory for this class
void LiNd::setLimit () {}

//--------------------------------------------------------------------
//release the memory of this class
void LiNd::release () {
	exDesNdNoLs_.release ();
	//int nPtn_ = ptnlist.getNPtn ();
	//int nRate_ = myrate.getNRate ();
	//int nState_ = mymodel.getNState ();
	Utl::delMem (liNdCube_);
//	Utl::delMem (liscale, nRate_);
}

//-------------------------------------------------------------------
//the destructor function
LiNd::~LiNd () {
	release ();
	//  std::cout << "this is the destructor of liInNd class " << endl;
}
